import json
import shutil

import cv2
import numpy as np
from lib.app.exceptions import AppException
from lib.core import DEPRICATED_DOCUMENT_VIDEO_MESSAGE
from superannotate.logger import get_default_logger

from ..common import blue_color_generator
from ..common import hex_to_rgb
from ..common import write_to_json

logger = get_default_logger()


def copy_file(src_path, dst_path):
    shutil.copy(src_path, dst_path)


def from_pixel_to_vector(json_paths, output_dir):
    img_names = []
    for json_path in json_paths:
        file_name = str(json_path.name).replace("___pixel.json", "___objects.json")

        mask_name = str(json_path).replace("___pixel.json", "___save.png")
        img = cv2.imread(mask_name)
        H, W, _ = img.shape

        sa_json = json.load(open(json_path))
        instances = sa_json["instances"]
        idx = 0
        sa_instances = []
        for instance in instances:
            if "parts" not in instance.keys():
                if "type" in instance.keys() and instance["type"] == "meta":
                    sa_instances.append(instance)
                continue

            parts = instance["parts"]

            polygons = []
            for part in parts:
                color = list(hex_to_rgb(part["color"]))
                mask = np.zeros((H, W), dtype=np.uint8)
                mask[np.all((img == color[::-1]), axis=2)] = 255
                contours, _ = cv2.findContours(
                    mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
                )
                part_polygons = []
                for contour in contours:
                    segment = contour.flatten().tolist()
                    if len(segment) > 6:
                        part_polygons.append(segment)
                polygons.append(part_polygons)

            for part_polygons in polygons:
                if len(part_polygons) > 1:
                    idx += 1
                    group_id = idx
                else:
                    group_id = 0

                for polygon in part_polygons:
                    temp = instance.copy()
                    del temp["parts"]
                    temp["pointLabels"] = {}
                    temp["groupId"] = group_id
                    temp["type"] = "polygon"
                    temp["points"] = polygon
                    sa_instances.append(temp.copy())
                    temp["type"] = "bbox"
                    temp["points"] = {
                        "x1": min(polygon[::2]),
                        "x2": max(polygon[::2]),
                        "y1": min(polygon[1::2]),
                        "y2": max(polygon[1::2]),
                    }
                    sa_instances.append(temp.copy())

        sa_json["instances"] = sa_instances
        write_to_json(output_dir / file_name, sa_json)
        img_names.append(file_name.replace("___objects.json", ""))
    return img_names


def from_vector_to_pixel(json_paths, output_dir):
    img_names = []
    for json_path in json_paths:
        file_name = str(json_path.name).replace("___objects.json", "___pixel.json")

        img_name = str(json_path).replace("___objects.json", "")
        img = cv2.imread(img_name)
        H, W, _ = img.shape

        sa_json = json.load(open(json_path))
        instances = sa_json["instances"]
        mask = np.zeros((H, W, 4))

        sa_instances = []
        blue_colors = blue_color_generator(len(instances))
        instances_group = {}
        for idx, instance in enumerate(instances):
            if instance["type"] == "polygon":
                if instance["groupId"] in instances_group.keys():
                    instances_group[instance["groupId"]].append(instance)
                else:
                    instances_group[instance["groupId"]] = [instance]
            elif instance["type"] == "meta":
                sa_instances.append(instance)

        idx = 0
        for key, instances in instances_group.items():
            if key == 0:
                for instance in instances:
                    pts = np.array(
                        [
                            instance["points"][2 * i : 2 * (i + 1)]
                            for i in range(len(instance["points"]) // 2)
                        ],
                        dtype=np.int32,
                    )
                    bitmask = np.zeros((H, W))
                    cv2.fillPoly(bitmask, [pts], 1)
                    mask[bitmask == 1] = list(hex_to_rgb(blue_colors[idx]))[::-1] + [
                        255
                    ]
                    del instance["type"]
                    del instance["points"]
                    del instance["pointLabels"]
                    del instance["groupId"]
                    instance["parts"] = [{"color": blue_colors[idx]}]
                    sa_instances.append(instance.copy())
                    idx += 1
            else:
                parts = []
                for instance in instances:
                    pts = np.array(
                        [
                            instance["points"][2 * i : 2 * (i + 1)]
                            for i in range(len(instance["points"]) // 2)
                        ],
                        dtype=np.int32,
                    )
                    bitmask = np.zeros((H, W))
                    cv2.fillPoly(bitmask, [pts], 1)
                    mask[bitmask == 1] = list(hex_to_rgb(blue_colors[idx]))[::-1] + [
                        255
                    ]
                    parts.append({"color": blue_colors[idx]})
                    idx += 1
                del instance["type"]
                del instance["points"]
                del instance["pointLabels"]
                del instance["groupId"]
                instance["parts"] = parts
                sa_instances.append(instance.copy())

        mask_name = file_name.replace("___pixel.json", "___save.png")
        cv2.imwrite(str(output_dir.joinpath(mask_name)), mask)

        sa_json["instances"] = sa_instances
        write_to_json(output_dir / file_name, sa_json)
        img_names.append(file_name.replace("___pixel.json", ""))
    return img_names


def sa_convert_project_type(input_dir, output_dir):
    json_paths = list(input_dir.glob("*.json"))

    output_dir.joinpath("classes").mkdir(parents=True)
    copy_file(
        input_dir.joinpath("classes", "classes.json"),
        output_dir.joinpath("classes", "classes.json"),
    )

    if "___pixel.json" in json_paths[0].name:
        img_names = from_pixel_to_vector(json_paths, output_dir)
    elif "___objects.json" in json_paths[0].name:
        img_names = from_vector_to_pixel(json_paths, output_dir)
    elif ".json" in json_paths[0].name:
        raise AppException(DEPRICATED_DOCUMENT_VIDEO_MESSAGE)
    else:
        raise AppException(
            "'input_dir' should contain JSON files with '[IMAGE_NAME]___objects.json' or '[IMAGE_NAME]___pixel.json' names structure.",
        )

    for img_name in img_names:
        copy_file(input_dir.joinpath(img_name), output_dir.joinpath(img_name))


def upgrade_json(input_dir, output_dir):
    files_list = list(input_dir.glob("*.json"))
    ptype = "Vector"
    if "___pixel" in str(files_list[0].name):
        ptype = "Pixel"

    converted_files = []
    failed_files = []
    for file in files_list:
        file_name = file.name
        try:
            output_json = _update_json_format(file, ptype)
            converted_files.append(file_name)
            write_to_json(output_dir / file_name, output_json)
        except Exception as e:
            logger.debug(str(e), exc_info=True)
            failed_files.append(file_name)

    return converted_files


def degrade_json(input_dir, output_dir):
    files_list = list(input_dir.glob("*.json"))

    converted_files = []
    failed_files = []
    for file in files_list:
        file_name = file.name
        try:
            output_json = _degrade_json_format(file)
            converted_files.append(output_dir / file_name)
            write_to_json(output_dir / file_name, output_json)
        except Exception as e:
            failed_files.append(file_name)

    return converted_files


def _update_json_format(old_json_path, project_type):
    old_json_data = json.load(open(old_json_path))
    new_json_data = {"metadata": {}, "instances": [], "tags": [], "comments": []}

    meta_keys = [
        "name",
        "width",
        "height",
        "status",
        "pinned",
        "isPredicted",
        "projectId",
        "annotatorEmail",
        "qaEmail",
    ]
    if project_type == "Pixel":
        meta_keys.append("isSegmented")

    new_json_data["metadata"] = dict.fromkeys(meta_keys)

    suffix = "___objects.json" if project_type == "Vector" else "___pixel.json"
    image_name = str(old_json_path.name).split(suffix)[0]
    metadata = new_json_data["metadata"]
    metadata["name"] = image_name

    for item in old_json_data:
        object_type = item.get("type")
        if object_type == "meta":
            meta_name = item["name"]
            if meta_name == "imageAttributes":
                metadata["height"] = item.get("height")
                metadata["width"] = item.get("width")
                metadata["status"] = item.get("status")
                metadata["pinned"] = item.get("pinned")
            if meta_name == "lastAction":
                metadata["lastAction"] = dict.fromkeys(["email", "timestamp"])
                metadata["lastAction"]["email"] = item.get("userId")
                metadata["lastAction"]["timestamp"] = item.get("timestamp")
        elif object_type == "tag":
            new_json_data["tags"].append(item.get("name"))
        elif object_type == "comment":
            item.pop("type")
            item["correspondence"] = item["comments"]
            for comment in item["correspondence"]:
                comment["email"] = comment["id"]
                comment.pop("id")
            item.pop("comments")
            new_json_data["comments"].append(item)
        else:
            new_json_data["instances"].append(item)

    return new_json_data


def _degrade_json_format(new_json_path):
    sa_loader = []
    new_json_data = json.load(open(new_json_path))

    meta = {"type": "meta", "name": "imageAttributes"}
    meta_keys = ["height", "width", "status", "pinned"]
    for meta_key in meta_keys:
        if meta_key in new_json_data["metadata"]:
            meta[meta_key] = new_json_data["metadata"][meta_key]
    sa_loader.append(meta)

    if "lastAction" in new_json_data["metadata"]:
        meta = {
            "type": "meta",
            "name": "lastAction",
            "userId": new_json_data["metadata"]["lastAction"]["email"],
            "timestamp": new_json_data["metadata"]["lastAction"]["timestamp"],
        }
        sa_loader.append(meta)

    for item in new_json_data["instances"]:
        sa_loader.append(item)

    for item in new_json_data["comments"]:
        comments = []
        for item2 in item["correspondence"]:
            comments.append({"text": item2["text"], "id": item2["email"]})
            item["comments"] = comments
        item["createdAt"] = item["correspondence"][0]["timestamp"]
        item["createdBy"] = {
            "email": item["correspondence"][0]["email"],
            "role": item["correspondence"][0]["role"],
        }
        item["updatedAt"] = item["correspondence"][-1]["timestamp"]
        item["updatedBy"] = {
            "email": item["correspondence"][-1]["email"],
            "role": item["correspondence"][-1]["role"],
        }
        item.pop("correspondence")
        item["type"] = "comment"
        item["comments"] = comments
        sa_loader.append(item)

    for item in new_json_data["tags"]:
        tag = {"type": "tag", "name": item}
        sa_loader.append(tag)

    return sa_loader



"""
"""
import numpy as np

from ....common import write_to_json
from ..baseStrategy import baseStrategy


class DataLoopStrategy(baseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def to_sa_format(self):
        classes = self.conversion_algorithm(
            self.export_root, self.task, self.output_dir
        )
        sa_classes = self._create_sa_classes(classes)
        (self.output_dir / "classes").mkdir(exist_ok=True)
        write_to_json(self.output_dir / "classes" / "classes.json", sa_classes)

    def _create_sa_classes(self, classes):
        sa_classes_loader = []
        for key, value in classes.items():
            color = np.random.choice(range(256), size=3)
            hexcolor = "#%02x%02x%02x" % tuple(color)
            sa_classes = {"name": key, "color": hexcolor, "attribute_groups": []}
            attribute_groups = []
            for attr_group_key, attr_group in value["attribute_group"].items():
                attr_loader = {
                    "name": attr_group_key,
                    "is_multiselect": attr_group["is_multiselect"],
                    "attributes": [],
                }
                for attr in set(attr_group["attributes"]):
                    attr_loader["attributes"].append({"name": attr})
                if attr_loader:
                    attribute_groups.append(attr_loader)

            sa_classes["attribute_groups"] = attribute_groups

            sa_classes_loader.append(sa_classes)
        return sa_classes_loader



import io
from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import cv2
import ffmpeg
from lib.core.exceptions import ImageProcessingException
from PIL import Image
from PIL import ImageDraw
from PIL import ImageOps
from superannotate.logger import get_default_logger

logger = get_default_logger()


class ImagePlugin:
    def __init__(self, image_bytes: io.BytesIO, max_resolution: int = 4096):
        self._image_bytes = image_bytes
        self._image_bytes.seek(0)
        self._max_resolution = max_resolution
        self._image = Image.open(self._image_bytes).convert("RGBA")
        self._draw = None

    def save(self, *args, **kwargs):
        self._image.save(*args, **kwargs)

    @staticmethod
    def from_array(arr):
        return Image.fromarray(arr)

    @staticmethod
    def Draw(image):
        return ImageDraw.Draw(image)

    @property
    def content(self):
        return self._image

    def show(self):
        self._image.show()

    def get_empty_image(self):
        return Image.new("RGBA", self._image.size)

    def get_empty(self):
        image_bytes = io.BytesIO()
        Image.new("RGB", self._image.size).save(image_bytes, "jpeg")
        image_bytes.seek(0)
        return ImagePlugin(image_bytes=image_bytes)

    @property
    def draw(self):
        if not self._draw:
            self._draw = ImageDraw.Draw(self._image)
        return self._draw

    def _get_image(self):
        Image.MAX_IMAGE_PIXELS = None
        im = self._image

        im = ImageOps.exif_transpose(im)

        width, height = im.size

        resolution = width * height

        if resolution > self._max_resolution:
            raise ImageProcessingException(
                f"Image resolution {resolution} too large. Max supported for resolution is {self._max_resolution}"
            )
        return im

    def get_size(self) -> Tuple[float, float]:
        return self._image.size

    def generate_thumb(self):
        image = self._get_image()
        buffer = io.BytesIO()

        thumbnail_size = (128, 96)
        background = Image.new("RGB", thumbnail_size, "black")
        image.thumbnail(thumbnail_size, Image.ANTIALIAS)
        (w, h) = image.size
        background.paste(
            image, ((thumbnail_size[0] - w) // 2, (thumbnail_size[1] - h) // 2)
        )
        im = background
        im.save(buffer, "JPEG")

        buffer.seek(0)
        width, height = im.size
        return buffer, width, height

    def generate_huge(self, base_width: int = 600) -> Tuple[io.BytesIO, float, float]:
        im = self._get_image()
        width, height = im.size
        buffer = io.BytesIO()
        h_size = int(height * base_width / width)
        im.resize((base_width, h_size), Image.ANTIALIAS).convert("RGB").save(
            buffer, "JPEG"
        )
        buffer.seek(0)
        width, height = im.size
        return buffer, width, height

    def generate_low_resolution(self, quality: int = 60, subsampling: int = -1):
        im = self._get_image()
        buffer = io.BytesIO()
        bg = Image.new("RGBA", im.size, (255, 255, 255))
        im = im.convert("RGBA")
        bg.paste(im, mask=im)
        bg = bg.convert("RGB")
        bg.save(buffer, "JPEG", quality=quality, subsampling=subsampling)
        buffer.seek(0)
        width, height = im.size
        return buffer, width, height

    def draw_bbox(self, x1, x2, y1, y2, fill_color, outline_color):
        image = self.get_empty_image()
        draw = ImageDraw.Draw(image)
        draw.rectangle(((x1, y1), (x2, y2)), fill_color, outline_color)
        self._image.convert("RGBA")
        self._image = Image.alpha_composite(self._image, image)

    def draw_polygon(self, points: List, fill_color, outline_color):
        image = self.get_empty_image()
        draw = ImageDraw.Draw(image)
        draw.polygon(points, fill_color, outline_color)
        self._image.convert("RGBA")
        self._image = Image.alpha_composite(self._image, image)

    def draw_polyline(self, points: List, fill_color, width=2):
        image = self.get_empty_image()
        draw = ImageDraw.Draw(image)
        draw.line(points, fill_color, width=width)
        self._image.convert("RGBA")
        self._image = Image.alpha_composite(self._image, image)

    def draw_point(self, x, y, fill_color, outline_color, size=2):
        image = self.get_empty_image()
        draw = ImageDraw.Draw(image)
        draw.ellipse(
            (x - size, y - size, x + size, y + size), fill_color, outline_color
        )
        self._image.convert("RGBA")
        self._image = Image.alpha_composite(self._image, image)

    def draw_ellipse(self, cx, cy, rx, ry, fill_color, outline_color, fixed=False):
        image = self.get_empty_image()
        draw = ImageDraw.Draw(image)
        if fixed:
            draw.ellipse((cx, cy, rx, ry), fill=fill_color, outline=outline_color)
        else:
            draw.ellipse(
                (cx - rx, cy - ry, cx + rx, cy + ry),
                fill=fill_color,
                outline=outline_color,
            )
        self._image.convert("RGBA")
        self._image = Image.alpha_composite(self._image, image)

    def draw_line(self, x, y, fill_color, width=1):
        image = self.get_empty_image()
        draw = ImageDraw.Draw(image)
        draw.line((x, y), fill=fill_color, width=width)
        self._image.convert("RGBA")
        self._image = Image.alpha_composite(self._image, image)


class VideoPlugin:
    @staticmethod
    def get_frames_count(video_path: str):
        video = cv2.VideoCapture(str(video_path), cv2.CAP_FFMPEG)
        count = 0
        flag = True
        while flag:
            flag, _ = video.read()
            if flag:
                count += 1
            else:
                break
        return count

    @staticmethod
    def get_fps(video_path: str) -> Union[int, float]:
        video = cv2.VideoCapture(str(video_path), cv2.CAP_FFMPEG)
        return video.get(cv2.CAP_PROP_FPS)

    @staticmethod
    def get_video_rotate_code(video_path, log):
        cv2_rotations = {
            90: cv2.ROTATE_90_CLOCKWISE,
            180: cv2.ROTATE_180,
            270: cv2.ROTATE_90_COUNTERCLOCKWISE,
        }
        try:
            meta_dict = ffmpeg.probe(str(video_path))
            rot = int(meta_dict["streams"][0]["tags"]["rotate"])
            if rot:
                if log:
                    logger.info(
                        "Frame rotation of %s found. Output images will be rotated accordingly.",
                        rot,
                    )
                return cv2_rotations[rot]
        except Exception as e:
            warning_str = ""
            if "ffprobe" in str(e):
                warning_str = "This could be because ffmpeg package is not installed. To install it, run: sudo apt install ffmpeg"
            if log:
                logger.warning(
                    "Couldn't read video metadata to determine rotation. %s",
                    warning_str,
                )
            return

    @staticmethod
    def frames_generator(
        video_path: str, start_time, end_time, target_fps: Optional[float], log=True
    ):
        video = cv2.VideoCapture(str(video_path), cv2.CAP_FFMPEG)
        if not video.isOpened():
            raise ImageProcessingException(
                f"Couldn't open video file {str(video_path)}."
            )
        fps = video.get(cv2.CAP_PROP_FPS)
        if not target_fps:
            target_fps = fps
        if target_fps > fps:
            target_fps = fps
        ratio = fps / target_fps
        rotate_code = VideoPlugin.get_video_rotate_code(video_path, log)
        frame_no = 0
        frame_no_with_change = 1.0
        while True:
            success, frame = video.read()
            if not success:
                break
            frame_no += 1
            if round(frame_no_with_change) != frame_no:
                continue
            frame_no_with_change += ratio
            frame_time = video.get(cv2.CAP_PROP_POS_MSEC) / 1000.0
            if end_time and frame_time > end_time:
                break
            if frame_time < start_time:
                continue
            if rotate_code:
                frame = cv2.rotate(frame, rotate_code)
            yield frame

    @staticmethod
    def get_extractable_frames(
        video_path: str, start_time, end_time, target_fps: float,
    ):
        total = VideoPlugin.get_frames_count(video_path)
        total_with_fps = sum(
            1
            for _ in VideoPlugin.frames_generator(
                video_path, start_time, end_time, target_fps, log=False
            )
        )
        zero_fill_count = len(str(total))
        video_name = Path(video_path).stem
        frame_names = []
        for i in range(1, total_with_fps + 1):
            frame_name = f"{video_name}_{str(i).zfill(zero_fill_count)}.jpg"
            frame_names.append(frame_name)
        return frame_names

    @staticmethod
    def extract_frames(
        video_path: str,
        start_time,
        end_time,
        extract_path: str,
        limit: int,
        target_fps: float,
        chunk_size: int = 100,
    ) -> List[str]:
        total_num_of_frames = VideoPlugin.get_frames_count(video_path)
        zero_fill_count = len(str(total_num_of_frames))
        video_name = Path(video_path).stem
        extracted_frame_no = 1
        extracted_frames_paths = []
        for frame in VideoPlugin.frames_generator(
            video_path, start_time, end_time, target_fps
        ):
            if len(extracted_frames_paths) >= limit:
                break
            path = str(
                Path(extract_path)
                / (
                    video_name
                    + "_"
                    + str(extracted_frame_no).zfill(zero_fill_count)
                    + ".jpg"
                )
            )
            extracted_frame_no += 1
            cv2.imwrite(path, frame)
            extracted_frames_paths.append(path)
            if len(extracted_frames_paths) % chunk_size == 0:
                yield extracted_frames_paths
                extracted_frames_paths.clear()
        if extracted_frames_paths:
            yield extracted_frames_paths



"""
VOC to SA conversion method
"""
import threading

import cv2
import numpy as np
from superannotate.logger import get_default_logger

from ....common import blue_color_generator
from ....common import hex_to_rgb
from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_pixel_instance
from ..sa_json_helper import _create_sa_json
from .voc_helper import _get_image_shape_from_xml
from .voc_helper import _get_voc_instances_from_xml
from .voc_helper import _iou

logger = get_default_logger()


def _generate_polygons(object_mask_path):
    segmentation = []

    object_mask = cv2.imread(str(object_mask_path), cv2.IMREAD_GRAYSCALE)

    object_unique_colors = np.unique(object_mask)

    num_colors = len([i for i in object_unique_colors if i not in (0, 220)])
    bluemask_colors = blue_color_generator(num_colors)
    H, W = object_mask.shape
    sa_mask = np.zeros((H, W, 4))
    i = 0
    for unique_color in object_unique_colors:
        if unique_color in (0, 220):
            continue

        mask = np.zeros_like(object_mask)
        mask[object_mask == unique_color] = 255
        contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        segment = []
        for contour in contours:
            contour = contour.flatten().tolist()
            segment += contour
        if len(contour) > 4:
            segmentation.append(segment)
        if len(segmentation) == 0:
            continue

        sa_mask[object_mask == unique_color] = [255] + list(
            hex_to_rgb(bluemask_colors[i])
        )
        i += 1
    return segmentation, sa_mask, bluemask_colors


def _generate_instances(polygon_instances, voc_instances, bluemask_colors):
    instances = []
    i = 0
    for polygon in polygon_instances:
        ious = []
        bbox_poly = [
            min(polygon[::2]),
            min(polygon[1::2]),
            max(polygon[::2]),
            max(polygon[1::2]),
        ]
        for _, bbox in voc_instances:
            ious.append(_iou(bbox_poly, bbox))
        ind = np.argmax(ious)
        class_name = list(voc_instances[ind][0].keys())[0]
        attributes = voc_instances[ind][0][class_name]
        instances.append(
            {
                "className": class_name,
                "polygon": polygon,
                "bbox": voc_instances[ind][1],
                "blue_color": bluemask_colors[i],
                "classAttributes": attributes,
            }
        )
        i += 1
    return instances


def voc_instance_segmentation_to_sa_pixel(voc_root, output_dir):
    classes = []
    object_masks_dir = voc_root / "SegmentationObject"
    annotation_dir = voc_root / "Annotations"

    file_list = list(object_masks_dir.glob("*"))
    if not file_list:
        logger.warning(
            "You need to have both 'Annotations' and 'SegmentationObject' directories to be able to convert."
        )
    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(file_list), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for filename in file_list:
        polygon_instances, sa_mask, bluemask_colors = _generate_polygons(
            object_masks_dir / filename.name
        )
        voc_instances = _get_voc_instances_from_xml(annotation_dir / filename.name)
        for class_, _ in voc_instances:
            classes.append(class_)

        maped_instances = _generate_instances(
            polygon_instances, voc_instances, bluemask_colors
        )

        sa_instances = []
        for instance in maped_instances:
            parts = [{"color": instance["blue_color"]}]
            sa_obj = _create_pixel_instance(
                parts, instance["classAttributes"], instance["className"]
            )
            sa_instances.append(sa_obj)

        images_converted.append(filename)
        file_name = "%s.jpg___pixel.json" % (filename.stem)
        height, width = _get_image_shape_from_xml(annotation_dir / filename.name)
        sa_metadata = {"name": filename.stem, "height": height, "width": width}
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)

        mask_name = "%s.jpg___save.png" % (filename.stem)
        cv2.imwrite(str(output_dir / mask_name), sa_mask[:, :, ::-1])
    finish_event.set()
    tqdm_thread.join()
    return classes



"""
"""
import json
from collections import namedtuple
from datetime import datetime
from pathlib import Path

import cv2
import numpy as np
from PIL import Image
from superannotate.logger import get_default_logger

from ....common import id2rgb
from ....common import write_to_json
from ..baseStrategy import baseStrategy

logger = get_default_logger()


class CocoBaseStrategy(baseStrategy):
    project_type_to_json_ending = {
        "pixel": "___pixel.json",
        "vector": "___objects.json",
    }

    def __init__(self, args):
        self.total_images_num = 0
        super().__init__(args)

    def set_num_total_images(self, num):
        self.total_images_num = num

    def get_num_total_images(self):
        return self.total_images_num

    def _create_categories(self, path_to_classes):

        classes = None
        s_class = namedtuple("Class", ["class_name", "id"])

        with open(path_to_classes) as fp:
            classes = json.load(fp)
        categories = [
            self._create_single_category(s_class(item, classes[item]))
            for item in classes
        ]
        return categories

    def _create_single_category(self, item):
        category = {
            "id": item.id,
            "name": item.class_name,
            "supercategory": item.class_name,
            "isthing": 1,
            "color": id2rgb(item.id),
        }
        return category

    def _make_id_generator(self):
        cur_id = 0
        while True:
            cur_id += 1
            yield cur_id

    def _create_skeleton(self):
        out_json = {
            "info": {
                "description": f"This is {self.dataset_name} dataset.",
                "url": "https://superannotate.ai",
                "version": "1.0",
                "year": datetime.now().year,
                "contributor": "Superannotate AI",
                "date_created": datetime.now().strftime("%d/%m/%Y"),
            },
            "licenses": [
                {"url": "https://superannotate.ai", "id": 1, "name": "Superannotate AI"}
            ],
            "images": [],
            "annotations": [],
            "categories": [],
        }
        return out_json

    def convert_from_old_sa_to_new(self, old_json_data, project_type):

        new_json_data = {"metadata": {}, "instances": [], "tags": [], "comments": []}

        meta_keys = [
            "name",
            "width",
            "height",
            "status",
            "pinned",
            "isPredicted",
            "projectId",
            "annotatorEmail",
            "qaEmail",
        ]
        if project_type == "pixel":
            meta_keys.append("isSegmented")

        new_json_data["metadata"] = dict.fromkeys(meta_keys)

        metadata = new_json_data["metadata"]

        for item in old_json_data:
            object_type = item.get("type")

            # add metadata
            if object_type == "meta":
                meta_name = item["name"]
                if meta_name == "imageAttributes":
                    metadata["height"] = item.get("height")
                    metadata["width"] = item.get("width")
                    metadata["status"] = item.get("status")
                    metadata["pinned"] = item.get("pinned")
                if meta_name == "lastAction":
                    metadata["lastAction"] = dict.fromkeys(["email", "timestamp"])
                    metadata["lastAction"]["email"] = item.get("userId")
                    metadata["lastAction"]["timestamp"] = item.get("timestamp")
            # add tags
            elif object_type == "tag":
                new_json_data["tags"].append(item.get("name"))
            # add comments
            elif object_type == "comment":
                item.pop("type")
                item["correspondence"] = item["comments"]
                for comment in item["correspondence"]:
                    comment["email"] = comment["id"]
                    comment.pop("id")
                item.pop("comments")
                new_json_data["comments"].append(item)
            # add instances
            else:
                new_json_data["instances"].append(item)
        return new_json_data

    def _parse_json_into_common_format(self, sa_annotation_json, fpath):
        """
           If the annotation format ever changes this function will handle it and
           return something optimal for the converters. Additionally, if anything
           important is absent from the current json, this function fills it.
        """
        if isinstance(sa_annotation_json, list):
            sa_annotation_json = self.convert_from_old_sa_to_new(
                sa_annotation_json, self.project_type
            )
        if "metadata" not in sa_annotation_json:
            sa_annotation_json["metadata"] = {}

        if "tags" not in sa_annotation_json:
            sa_annotation_json["tags"] = []

        if "instances" not in sa_annotation_json:
            sa_annotation_json["instances"] = []
        if "comments" not in sa_annotation_json:
            sa_annotation_json["comments"] = []

        if (
            "name" not in sa_annotation_json["metadata"]
            or sa_annotation_json["metadata"]["name"] is None
        ):
            fname = fpath.name
            fname = fname[
                : -len(self.project_type_to_json_ending[self.project_type.lower()])
            ]
            sa_annotation_json["metadata"]["name"] = fname
        sa_annotation_json["metadata"]["image_path"] = str(
            Path(fpath).parent / sa_annotation_json["metadata"]["name"]
        )

        sa_annotation_json["metadata"]["annotation_json"] = fpath
        if self.task == "panoptic_segmentation":
            panoptic_mask = str(
                Path(self.export_root)
                / (sa_annotation_json["metadata"]["name"] + ".png")
            )

            sa_annotation_json["metadata"]["panoptic_mask"] = panoptic_mask

        if self.project_type == "Pixel":
            sa_annotation_json["metadata"]["sa_bluemask_path"] = str(
                Path(self.export_root)
                / (sa_annotation_json["metadata"]["name"] + "___save.png")
            )

        if not isinstance(
            sa_annotation_json["metadata"].get("height", None), int
        ) or not isinstance(sa_annotation_json["metadata"].get("width", None), int):
            image_height, image_width = self.get_image_dimensions(
                sa_annotation_json["metadata"]["image_path"]
            )
            sa_annotation_json["metadata"]["height"] = image_height
            sa_annotation_json["metadata"]["width"] = image_width

        return sa_annotation_json

    def get_image_dimensions(self, image_path):

        img_height = None
        img_width = None

        img = cv2.imread(image_path, cv2.IMREAD_UNCHANGED)
        if img is not None:
            dimensions = img.shape
            img_height, img_width = (dimensions[0], dimensions[1])
        else:
            try:
                img = Image.open(image_path)
                img_width, img_height = img.size()
            except Exception as e:
                raise

        return img_height, img_width

    def _prepare_single_image_commons_pixel(self, id_, metadata):

        ImgCommons = namedtuple(
            "ImgCommons", ["image_info", "ann_mask", "sa_bluemask_rgb", "flat_mask"]
        )
        sa_bluemask_path = metadata["sa_bluemask_path"]

        image_info = self._make_image_info(
            metadata["name"], metadata["height"], metadata["width"], id_
        )

        sa_bluemask_rgb = np.asarray(
            Image.open(sa_bluemask_path).convert("RGB"), dtype=np.uint32
        )

        ann_mask = np.zeros(
            (image_info["height"], image_info["width"]), dtype=np.uint32
        )
        flat_mask = (
            (sa_bluemask_rgb[:, :, 0] << 16)
            | (sa_bluemask_rgb[:, :, 1] << 8)
            | (sa_bluemask_rgb[:, :, 2])
        )

        res = ImgCommons(image_info, ann_mask, sa_bluemask_rgb, flat_mask)

        return res

    def _prepare_single_image_commons_vector(self, id_, metadata):

        ImgCommons = namedtuple("ImgCommons", ["image_info"])

        image_info = self._make_image_info(
            metadata["name"], metadata["height"], metadata["width"], id_
        )

        res = ImgCommons(image_info)

        return res

    def _prepare_single_image_commons(self, id_, metadata):
        res = None
        if self.project_type == "Pixel":
            res = self._prepare_single_image_commons_pixel(id_, metadata)
        elif self.project_type == "Vector":
            res = self._prepare_single_image_commons_vector(id_, metadata)
        return res

    def _make_image_info(self, pname, pheight, pwidth, id_):
        image_info = {
            "id": id_,
            "file_name": pname,
            "height": pheight,
            "width": pwidth,
            "license": 1,
        }

        return image_info

    def _create_sa_classes(self, json_path):
        json_data = json.load(open(json_path))
        classes_list = json_data["categories"]

        classes = []
        for data in classes_list:
            color = np.random.choice(range(256), size=3)
            hexcolor = "#%02x%02x%02x" % tuple(color)
            classes_dict = {
                "name": data["name"],
                "color": hexcolor,
                "attribute_groups": [],
            }
            classes.append(classes_dict)

        return classes

    def to_sa_format(self):
        json_data = self.export_root / (self.dataset_name + ".json")
        sa_classes = self._create_sa_classes(json_data)
        (self.output_dir / "classes").mkdir(parents=True, exist_ok=True)
        write_to_json(self.output_dir / "classes" / "classes.json", sa_classes)
        self.conversion_algorithm(json_data, self.output_dir)

    def make_anno_json_generator(self):
        json_data = None

        if self.project_type == "Pixel":
            jsons = list(Path(self.export_root).glob("*pixel.json"))
        elif self.project_type == "Vector":
            jsons = list(Path(self.export_root).glob("*objects.json"))

        self.set_num_total_images(len(jsons))
        print()
        for fpath in jsons:
            with open(fpath) as fp:
                json_data = json.load(fp)
                json_data = self._parse_json_into_common_format(json_data, fpath)

            yield json_data



from abc import ABC
from abc import ABCMeta
from abc import abstractmethod
from typing import Any
from typing import List
from typing import Optional
from typing import Union

import boto3
from lib.core.conditions import Condition
from lib.core.entities import BaseEntity
from lib.core.entities import Entity
from lib.core.entities import ProjectEntity
from lib.core.entities import TmpBaseEntity
from lib.core.serviceproviders import SuperannotateServiceProvider


class BaseReadOnlyRepository(ABC):
    @abstractmethod
    def get_one(
        self, uuid: Union[Condition, int]
    ) -> Optional[Union[BaseEntity, Entity]]:
        raise NotImplementedError

    @abstractmethod
    def get_all(
        self, condition: Optional[Condition] = None
    ) -> List[Union[BaseEntity, Entity]]:
        raise NotImplementedError

    @staticmethod
    def dict2entity(data: dict) -> Entity:
        return Entity(**TmpBaseEntity(**data).dict())


class BaseManageableRepository(BaseReadOnlyRepository):
    @abstractmethod
    def insert(self, entity: BaseEntity) -> BaseEntity:
        raise NotImplementedError

    @abstractmethod
    def update(self, entity: BaseEntity) -> BaseEntity:
        raise NotImplementedError

    @abstractmethod
    def delete(self, uuid: Any):
        raise NotImplementedError

    def bulk_delete(self, entities: List[BaseEntity]) -> bool:
        raise NotImplementedError

    @staticmethod
    def _drop_nones(data: dict):
        for k, v in list(data.items()):
            if v is None:
                del data[k]
        return data


class BaseProjectRelatedManageableRepository(BaseManageableRepository, metaclass=ABCMeta):
    def __init__(self, service: SuperannotateServiceProvider, project: ProjectEntity):
        self._service = service
        self._project = project


class BaseS3Repository(BaseManageableRepository):
    def __init__(
        self, access_key: str, secret_key: str, session_token: str, bucket: str,
    ):
        self._session = boto3.Session(
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            aws_session_token=session_token,
        )

        self._resource = self._session.resource("s3")
        self._bucket = bucket
        self.bucket = self._resource.Bucket(bucket)



import logging.config
import os
import sys

import requests
import superannotate.lib.core as constances
from packaging.version import parse
from superannotate.lib import get_default_controller
from superannotate.lib.app.analytics.class_analytics import class_distribution
from superannotate.lib.app.exceptions import AppException
from superannotate.lib.app.input_converters.conversion import convert_json_version
from superannotate.lib.app.input_converters.conversion import convert_project_type
from superannotate.lib.app.input_converters.conversion import export_annotation
from superannotate.lib.app.input_converters.conversion import import_annotation
from superannotate.lib.app.interface.sdk_interface import add_annotation_bbox_to_image
from superannotate.lib.app.interface.sdk_interface import (
    add_annotation_comment_to_image,
)
from superannotate.lib.app.interface.sdk_interface import add_annotation_point_to_image
from superannotate.lib.app.interface.sdk_interface import add_contributors_to_project
from superannotate.lib.app.interface.sdk_interface import aggregate_annotations_as_df
from superannotate.lib.app.interface.sdk_interface import assign_folder
from superannotate.lib.app.interface.sdk_interface import assign_images
from superannotate.lib.app.interface.sdk_interface import (
    attach_document_urls_to_project,
)
from superannotate.lib.app.interface.sdk_interface import attach_image_urls_to_project
from superannotate.lib.app.interface.sdk_interface import attach_items
from superannotate.lib.app.interface.sdk_interface import (
    attach_items_from_integrated_storage,
)
from superannotate.lib.app.interface.sdk_interface import attach_video_urls_to_project
from superannotate.lib.app.interface.sdk_interface import benchmark
from superannotate.lib.app.interface.sdk_interface import clone_project
from superannotate.lib.app.interface.sdk_interface import consensus
from superannotate.lib.app.interface.sdk_interface import copy_image
from superannotate.lib.app.interface.sdk_interface import copy_images
from superannotate.lib.app.interface.sdk_interface import copy_items
from superannotate.lib.app.interface.sdk_interface import create_annotation_class
from superannotate.lib.app.interface.sdk_interface import (
    create_annotation_classes_from_classes_json,
)
from superannotate.lib.app.interface.sdk_interface import create_folder
from superannotate.lib.app.interface.sdk_interface import create_project
from superannotate.lib.app.interface.sdk_interface import create_project_from_metadata
from superannotate.lib.app.interface.sdk_interface import delete_annotation_class
from superannotate.lib.app.interface.sdk_interface import delete_annotations
from superannotate.lib.app.interface.sdk_interface import delete_folders
from superannotate.lib.app.interface.sdk_interface import delete_images
from superannotate.lib.app.interface.sdk_interface import delete_project
from superannotate.lib.app.interface.sdk_interface import (
    download_annotation_classes_json,
)
from superannotate.lib.app.interface.sdk_interface import download_export
from superannotate.lib.app.interface.sdk_interface import download_image
from superannotate.lib.app.interface.sdk_interface import download_image_annotations
from superannotate.lib.app.interface.sdk_interface import download_model
from superannotate.lib.app.interface.sdk_interface import get_annotations
from superannotate.lib.app.interface.sdk_interface import get_annotations_per_frame
from superannotate.lib.app.interface.sdk_interface import get_exports
from superannotate.lib.app.interface.sdk_interface import get_folder_metadata
from superannotate.lib.app.interface.sdk_interface import get_image_metadata
from superannotate.lib.app.interface.sdk_interface import get_integrations
from superannotate.lib.app.interface.sdk_interface import get_item_metadata
from superannotate.lib.app.interface.sdk_interface import get_project_image_count
from superannotate.lib.app.interface.sdk_interface import get_project_metadata
from superannotate.lib.app.interface.sdk_interface import get_project_settings
from superannotate.lib.app.interface.sdk_interface import get_project_workflow
from superannotate.lib.app.interface.sdk_interface import get_team_metadata
from superannotate.lib.app.interface.sdk_interface import init
from superannotate.lib.app.interface.sdk_interface import invite_contributors_to_team
from superannotate.lib.app.interface.sdk_interface import move_images
from superannotate.lib.app.interface.sdk_interface import move_items
from superannotate.lib.app.interface.sdk_interface import pin_image
from superannotate.lib.app.interface.sdk_interface import prepare_export
from superannotate.lib.app.interface.sdk_interface import query
from superannotate.lib.app.interface.sdk_interface import rename_project
from superannotate.lib.app.interface.sdk_interface import run_prediction
from superannotate.lib.app.interface.sdk_interface import search_annotation_classes
from superannotate.lib.app.interface.sdk_interface import search_folders
from superannotate.lib.app.interface.sdk_interface import search_images
from superannotate.lib.app.interface.sdk_interface import search_images_all_folders
from superannotate.lib.app.interface.sdk_interface import search_items
from superannotate.lib.app.interface.sdk_interface import search_models
from superannotate.lib.app.interface.sdk_interface import search_projects
from superannotate.lib.app.interface.sdk_interface import search_team_contributors
from superannotate.lib.app.interface.sdk_interface import set_auth_token
from superannotate.lib.app.interface.sdk_interface import set_image_annotation_status
from superannotate.lib.app.interface.sdk_interface import set_images_annotation_statuses
from superannotate.lib.app.interface.sdk_interface import (
    set_project_default_image_quality_in_editor,
)
from superannotate.lib.app.interface.sdk_interface import set_project_workflow
from superannotate.lib.app.interface.sdk_interface import share_project
from superannotate.lib.app.interface.sdk_interface import unassign_folder
from superannotate.lib.app.interface.sdk_interface import unassign_images
from superannotate.lib.app.interface.sdk_interface import (
    upload_annotations_from_folder_to_project,
)
from superannotate.lib.app.interface.sdk_interface import upload_image_annotations
from superannotate.lib.app.interface.sdk_interface import upload_image_to_project
from superannotate.lib.app.interface.sdk_interface import (
    upload_images_from_folder_to_project,
)
from superannotate.lib.app.interface.sdk_interface import upload_images_to_project
from superannotate.lib.app.interface.sdk_interface import (
    upload_preannotations_from_folder_to_project,
)
from superannotate.lib.app.interface.sdk_interface import upload_priority_scores
from superannotate.lib.app.interface.sdk_interface import upload_video_to_project
from superannotate.lib.app.interface.sdk_interface import (
    upload_videos_from_folder_to_project,
)
from superannotate.lib.app.interface.sdk_interface import validate_annotations
from superannotate.logger import get_default_logger
from superannotate.version import __version__


controller = get_default_controller()


__all__ = [
    "__version__",
    "controller",
    "constances",
    # Utils
    "AppException",
    "validate_annotations",
    #
    "init",
    "set_auth_token",
    # analytics
    "class_distribution",
    "aggregate_annotations_as_df",
    "get_exports",
    # annotations
    "get_annotations",
    "get_annotations_per_frame",
    # integrations
    "get_integrations",
    "attach_items_from_integrated_storage",
    # converters
    "convert_json_version",
    "import_annotation",
    "export_annotation",
    "convert_project_type",
    # Teams Section
    "get_team_metadata",
    "search_team_contributors",
    # Projects Section
    "create_project_from_metadata",
    "get_project_settings",
    "get_project_metadata",
    "get_project_workflow",
    "set_project_workflow",
    "search_projects",
    "create_project",
    "clone_project",
    "share_project",
    "delete_project",
    "rename_project",
    "upload_priority_scores",
    # Images Section
    "search_images",
    "copy_image",
    # Folders Section
    "create_folder",
    "get_folder_metadata",
    "delete_folders",
    "search_folders",
    "assign_folder",
    "unassign_folder",
    # Items Section
    "get_item_metadata",
    "search_items",
    "query",
    "attach_items",
    "copy_items",
    "move_items",
    # Image Section
    "copy_images",
    "move_images",
    "delete_images",
    "download_image",
    "pin_image",
    "get_image_metadata",
    "get_project_image_count",
    "search_images_all_folders",
    "assign_images",
    "unassign_images",
    "download_image_annotations",
    "delete_annotations",
    "upload_image_to_project",
    "upload_image_annotations",
    "upload_images_from_folder_to_project",
    "attach_image_urls_to_project",
    "attach_video_urls_to_project",
    "attach_document_urls_to_project",
    # Video Section
    "upload_videos_from_folder_to_project",
    # Annotation Section
    "create_annotation_class",
    "delete_annotation_class",
    "prepare_export",
    "download_export",
    "set_images_annotation_statuses",
    "add_annotation_bbox_to_image",
    "add_annotation_point_to_image",
    "add_annotation_comment_to_image",
    "search_annotation_classes",
    "create_annotation_classes_from_classes_json",
    "upload_annotations_from_folder_to_project",
    "upload_preannotations_from_folder_to_project",
    "download_annotation_classes_json",
    "set_project_default_image_quality_in_editor",
    "run_prediction",
    "search_models",
    "download_model",
    "set_image_annotation_status",
    "benchmark",
    "consensus",
    "upload_video_to_project",
    "upload_images_to_project",
    "add_contributors_to_project",
    "invite_contributors_to_team",
]

__author__ = "Superannotate"

WORKING_DIR = os.path.split(os.path.realpath(__file__))[0]
sys.path.append(WORKING_DIR)
logging.getLogger("botocore").setLevel(logging.CRITICAL)
logger = get_default_logger()


def log_version_info():
    local_version = parse(__version__)
    if local_version.is_prerelease:
        logger.info(constances.PACKAGE_VERSION_INFO_MESSAGE.format(__version__))
    req = requests.get("https://pypi.python.org/pypi/superannotate/json")
    if req.ok:
        releases = req.json().get("releases", [])
        pip_version = parse("0")
        for release in releases:
            ver = parse(release)
            if not ver.is_prerelease or local_version.is_prerelease:
                pip_version = max(pip_version, ver)
        if pip_version.major > local_version.major:
            logger.warning(
                constances.PACKAGE_VERSION_MAJOR_UPGRADE.format(
                    local_version, pip_version
                )
            )
        elif pip_version > local_version:
            logger.warning(
                constances.PACKAGE_VERSION_UPGRADE.format(local_version, pip_version)
            )


if not os.environ.get("SA_VERSION_CHECK", "True").lower() == "false":
    log_version_info()



from datetime import datetime
from datetime import timedelta
from functools import lru_cache
from functools import wraps


def timed_lru_cache(seconds: int, maxsize: int = 32):
    def wrapper_cache(func):
        func = lru_cache(maxsize=maxsize)(func)
        func.lifetime = timedelta(seconds=seconds)
        func.expiration = datetime.utcnow() + func.lifetime

        @wraps(func)
        def wrapped_func(*args, **kwargs):
            wrapped_func.cache_clear = func.cache_clear
            if datetime.utcnow() >= func.expiration:
                func.cache_clear()
                func.expiration = datetime.utcnow() + func.lifetime

            return func(*args, **kwargs)

        return wrapped_func

    return wrapper_cache



from typing import Union


class Response:
    def __init__(self, status: str = None, data: Union[dict, list] = None):
        self._status = status
        self._data = data
        self._report = []
        self._errors = []

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def report(self):
        return "\n".join(self._report)

    @report.setter
    def report(self, value: str):
        self._report.append(value)

    @property
    def report_messages(self):
        return self._report

    @property
    def status(self):
        return self.data

    @status.setter
    def status(self, value):
        self._status = value

    @property
    def errors(self):
        return "\n".join([str(error) for error in self._errors])

    @errors.setter
    def errors(self, error: list):
        self._errors.append(error)



import numpy as np

from ....common import write_to_json
from ..baseStrategy import baseStrategy


class VGGStrategy(baseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def to_sa_format(self):
        json_data = self.export_root / (self.dataset_name + ".json")
        classes = self.conversion_algorithm(json_data, self.task, self.output_dir)
        sa_classes = self._create_classes(classes)
        (self.output_dir / "classes").mkdir(exist_ok=True)
        write_to_json(self.output_dir / "classes" / "classes.json", sa_classes)

    def _create_classes(self, class_id_map):
        sa_classes = []
        for key in class_id_map.keys():
            color = np.random.choice(range(256), size=3)
            hexcolor = "#%02x%02x%02x" % tuple(color)
            dd = {"name": key, "color": hexcolor, "attribute_groups": []}
            for attributes, value in class_id_map[key]["attribute_groups"].items():
                attr_group = {
                    "name": attributes,
                    "is_multiselect": value["is_multiselect"],
                    "attributes": [],
                }
                for attribute in value["attributes"]:
                    attr = {"name": attribute, "groupName": attributes}
                    attr_group["attributes"].append(attr.copy())
                dd["attribute_groups"].append(attr_group.copy())
            sa_classes.append(dd)
        return sa_classes



import cv2 as cv
import numpy as np

from .coco_api import _area
from .coco_api import _toBbox


def __instance_object_commons_per_instance(instance, id_generator, flat_mask):
    if "parts" not in instance:
        return None

    anno_id = next(id_generator)
    parts = [int(part["color"][1:], 16) for part in instance["parts"]]
    category_id = instance["classId"]

    instance_bitmask = np.isin(flat_mask, parts)

    databytes = instance_bitmask * np.uint8(255)
    contours, _ = cv.findContours(databytes, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE)
    bbox = list(_toBbox(instance_bitmask))
    area = int(_area(instance_bitmask.astype(np.uint8)))
    return (bbox, area, contours, category_id, anno_id)


def instance_object_commons(instances, id_generator, flat_mask):
    commons_lst = [
        __instance_object_commons_per_instance(x, id_generator, flat_mask)
        for x in instances
    ]
    commons_lst = [x for x in commons_lst if x is not None]
    return commons_lst


def sa_pixel_to_coco_instance_segmentation(
    make_annotation, image_commons, instances, id_generator
):
    commons_lst = instance_object_commons(
        instances, id_generator, image_commons.flat_mask
    )
    image_info = image_commons.image_info
    annotations_per_image = []
    for common in commons_lst:
        bbox, area, contours, category_id, anno_id = common
        segmentation = [
            contour.flatten().tolist()
            for contour in contours
            if len(contour.flatten().tolist()) >= 5
        ]

        if segmentation != []:
            annotations_per_image.append(
                make_annotation(
                    category_id, image_info["id"], bbox, segmentation, area, anno_id
                )
            )

    return (image_info, annotations_per_image)


def sa_pixel_to_coco_panoptic_segmentation(image_commons, instnaces, id_generator):
    flat_mask = image_commons.flat_mask
    ann_mask = image_commons.ann_mask

    segments_info = []

    for instance in instnaces:
        if "parts" not in instance:
            continue

        parts = [int(part["color"][1:], 16) for part in instance["parts"]]
        category_id = instance["classId"]
        instance_bitmask = np.isin(flat_mask, parts)
        segment_id = next(id_generator)
        ann_mask[instance_bitmask] = segment_id
        bbox = list(_toBbox(instance_bitmask))
        area = int(_area(instance_bitmask.astype(np.uint8)))

        segment_info = {
            "id": segment_id,
            "category_id": category_id,
            "area": area,
            "bbox": bbox,
            "iscrowd": 0,
        }

        segments_info.append(segment_info)

    return (image_commons.image_info, segments_info, image_commons.ann_mask)



import asyncio
import datetime
import json
import time
from contextlib import contextmanager
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Union
from urllib.parse import urljoin

import lib.core as constance
import requests.packages.urllib3
from lib.core.exceptions import AppException
from lib.core.exceptions import BackendError
from lib.core.reporter import Reporter
from lib.core.service_types import DownloadMLModelAuthData
from lib.core.service_types import ServiceResponse
from lib.core.service_types import UploadAnnotationAuthData
from lib.core.service_types import UserLimits
from lib.core.serviceproviders import SuperannotateServiceProvider
from lib.infrastructure.helpers import timed_lru_cache
from lib.infrastructure.stream_data_handler import StreamedAnnotations
from requests.exceptions import HTTPError

requests.packages.urllib3.disable_warnings()


class PydanticEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "deserialize"):
            return obj.deserialize()
        if isinstance(obj, (datetime.date, datetime.datetime)):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


class BaseBackendService(SuperannotateServiceProvider):
    AUTH_TYPE = "sdk"
    PAGINATE_BY = 100
    LIMIT = 100
    MAX_ITEMS_COUNT = 50 * 1000

    """
    Base service class
    """

    def __init__(
            self,
            api_url: str,
            auth_token: str,
            logger,
            paginate_by=None,
            verify_ssl=False,
            testing: bool = False,
    ):
        self.api_url = api_url
        self._auth_token = auth_token
        self.logger = logger
        self._paginate_by = paginate_by
        self._verify_ssl = verify_ssl
        self.team_id = auth_token.split("=")[-1]
        self._testing = testing
        self.get_session()

    @property
    def assets_provider_url(self):
        if self.api_url != constance.BACKEND_URL:
            return "https://assets-provider.devsuperannotate.com/api/v1/"
        return "https://assets-provider.superannotate.com/api/v1/"

    @timed_lru_cache(seconds=360)
    def get_session(self):
        session = requests.Session()
        session.headers.update(self.default_headers)
        return session

    @property
    def default_headers(self):
        return {
            "Authorization": self._auth_token,
            "authtype": self.AUTH_TYPE,
            "Content-Type": "application/json",
            # "User-Agent": constance.__version__,
        }

    @property
    def safe_api(self):
        """
        Context manager which will handle requests calls.
        """

        @contextmanager
        def safe_api():
            """
            Context manager which handles Requests error.
            """
            try:
                yield None
            except (HTTPError, ConnectionError) as exc:
                raise AppException(f"Unknown exception: {exc}.")

        return safe_api

    @property
    def paginate_by(self):
        if self._paginate_by:
            return self._paginate_by
        else:
            return self.PAGINATE_BY

    def _request(
            self,
            url,
            method="get",
            data=None,
            headers=None,
            params=None,
            retried=0,
            content_type=None,
    ) -> Union[requests.Response, ServiceResponse]:
        kwargs = {"data": json.dumps(data, cls=PydanticEncoder)} if data else {}
        session = self.get_session()
        session.headers.update(headers if headers else {})
        with self.safe_api():
            req = requests.Request(method=method, url=url, **kwargs, params=params)
            prepared = session.prepare_request(req)
            response = session.send(request=prepared, verify=self._verify_ssl)
        if response.status_code == 404 and retried < 3:
            return self._request(
                url,
                method="get",
                data=None,
                headers=None,
                params=None,
                retried=retried + 1,
                content_type=content_type,
            )
        if response.status_code > 299:
            self.logger.error(
                f"Got {response.status_code} response from backend: {response.text}"
            )
        if content_type:
            return ServiceResponse(response, content_type)
        return response

    def _get_page(self, url, offset, params=None, key_field: str = None):
        splitter = "&" if "?" in url else "?"
        url = f"{url}{splitter}offset={offset}"

        response = self._request(url, params=params)
        if response.status_code != 200:
            return {"data": []}, 0
            # raise AppException(f"Got invalid response for url {url}: {response.text}.")
        data = response.json()
        if data:
            if isinstance(data, dict):
                if key_field:
                    data = data[key_field]
                if data.get("count", 0) < self.LIMIT:
                    return data, 0
                else:
                    return data, data.get("count", 0) - offset
            if isinstance(data, list):
                return {"data": data}, 0
        return {"data": []}, 0

    def _get_all_pages(self, url, offset=0, params=None, key_field: str = None):
        total = list()

        while True:
            resources, remains_count = self._get_page(url, offset, params, key_field)
            total.extend(resources["data"])
            if remains_count <= 0:
                break
            offset += len(resources["data"])
        return total


class SuperannotateBackendService(BaseBackendService):
    """
    Manage projects, images and team in the Superannotate
    """

    DEFAULT_CHUNK_SIZE = 1000

    URL_USERS = "users"
    URL_LIST_PROJECTS = "projects"
    URL_FOLDERS_IMAGES = "images-folders"
    URL_CREATE_PROJECT = "project"
    URL_GET_PROJECT = "project/{}"
    URL_GET_FOLDER_BY_NAME = "folder/getFolderByName"
    URL_CREATE_FOLDER = "folder"
    URL_UPDATE_FOLDER = "folder/{}"
    URL_GET_IMAGE = "image/{}"
    URL_GET_IMAGES = "images"
    URL_GET_ITEMS = "items"
    URL_BULK_GET_IMAGES = "images/getBulk"
    URL_DELETE_FOLDERS = "image/delete/images"
    URL_CREATE_IMAGE = "image/ext-create"
    URL_PROJECT_SETTINGS = "project/{}/settings"
    URL_PROJECT_WORKFLOW = "project/{}/workflow"
    URL_SHARE_PROJECT = "project/{}/share"
    URL_SHARE_PROJECT_BULK = "project/{}/share/bulk"
    URL_ANNOTATION_CLASSES = "classes"
    URL_TEAM = "team"
    URL_INVITE_CONTRIBUTOR = "team/{}/invite"
    URL_INVITE_CONTRIBUTORS = "team/{}/inviteUsers"
    URL_PREPARE_EXPORT = "export"
    URL_COPY_IMAGES_FROM_FOLDER = "images/copy-image-or-folders"
    URL_MOVE_IMAGES_FROM_FOLDER = "image/move"
    URL_GET_COPY_PROGRESS = "images/copy-image-progress"
    URL_ASSIGN_IMAGES = "images/editAssignment/"
    URL_ASSIGN_FOLDER = "folder/editAssignment"
    URL_S3_ACCESS_POINT = "/project/{}/get-image-s3-access-point"
    URL_S3_UPLOAD_STATUS = "/project/{}/getS3UploadStatus"
    URL_GET_EXPORTS = "exports"
    URL_GET_CLASS = "class/{}"
    URL_ANNOTATION_UPLOAD_PATH_TOKEN = "images/getAnnotationsPathsAndTokens"
    URL_PRE_ANNOTATION_UPLOAD_PATH_TOKEN = "images/getPreAnnotationsPathsAndTokens"
    URL_GET_TEMPLATES = "templates"
    URL_PROJECT_WORKFLOW_ATTRIBUTE = "project/{}/workflow_attribute"
    URL_MODELS = "ml_models"
    URL_MODEL = "ml_model"
    URL_GET_MODEL_METRICS = "ml_models/{}/getCurrentMetrics"
    URL_BULK_GET_FOLDERS = "foldersByTeam"
    URL_GET_EXPORT = "export/{}"
    URL_GET_ML_MODEL_DOWNLOAD_TOKEN = "ml_model/getMyModelDownloadToken/{}"
    URL_PREDICTION = "images/prediction"
    URL_SET_IMAGES_STATUSES_BULK = "image/updateAnnotationStatusBulk"
    URL_DELETE_ANNOTATIONS = "annotations/remove"
    URL_DELETE_ANNOTATIONS_PROGRESS = "annotations/getRemoveStatus"
    URL_GET_LIMITS = "project/{}/limitationDetails"
    URL_GET_ANNOTATIONS = "images/annotations/stream"
    URL_UPLOAD_PRIORITY_SCORES = "images/updateEntropy"
    URL_GET_INTEGRATIONS = "integrations"
    URL_ATTACH_INTEGRATIONS = "image/integration/create"
    URL_SAQUL_QUERY = "/images/search/advanced"
    URL_VALIDATE_SAQUL_QUERY = "/images/parse/query/advanced"

    def upload_priority_scores(
            self, team_id: int, project_id: int, folder_id: int, priorities: list
    ) -> dict:
        upload_priority_score_url = urljoin(
            self.api_url, self.URL_UPLOAD_PRIORITY_SCORES
        )
        res = self._request(
            upload_priority_score_url,
            "post",
            params={
                "team_id": team_id,
                "project_id": project_id,
                "folder_id": folder_id,
            },
            data={"image_entropies": priorities},
        )
        return res.json()

    def get_project(self, uuid: int, team_id: int):
        get_project_url = urljoin(self.api_url, self.URL_GET_PROJECT.format(uuid))
        res = self._request(get_project_url, "get", params={"team_id": team_id})
        return res.json()

    def get_s3_upload_auth_token(self, team_id: int, folder_id: int, project_id: int):
        auth_token_url = urljoin(
            self.api_url,
            self.URL_GET_PROJECT.format(project_id) + "/sdkImageUploadToken",
        )
        response = self._request(
            auth_token_url, "get", params={"team_id": team_id, "folder_id": folder_id}
        )
        return response.json()

    def get_download_token(
            self,
            project_id: int,
            team_id: int,
            folder_id: int,
            image_id: int,
            include_original: int = 1,
    ):
        download_token_url = urljoin(
            self.api_url,
            self.URL_GET_IMAGE.format(image_id)
            + "/annotation/getAnnotationDownloadToken",
        )
        response = self._request(
            download_token_url,
            "get",
            params={
                "project_id": project_id,
                "team_id": team_id,
                "folder_id": folder_id,
                "include_original": include_original,
            },
        )
        return response.json()

    def get_upload_token(
            self, project_id: int, team_id: int, folder_id: int, image_id: int,
    ):
        download_token_url = urljoin(
            self.api_url,
            self.URL_GET_IMAGE.format(image_id)
            + "/annotation/getAnnotationUploadToken",
        )
        response = self._request(
            download_token_url,
            "get",
            params={
                "project_id": project_id,
                "team_id": team_id,
                "folder_id": folder_id,
            },
        )
        return response.json()

    def get_projects(self, query_string: str = None) -> list:
        url = urljoin(self.api_url, self.URL_LIST_PROJECTS)
        if query_string:
            url = f"{url}?{query_string}"
        data = self._get_all_pages(url)
        return data

    def create_project(self, project_data: dict) -> dict:
        create_project_url = urljoin(self.api_url, self.URL_CREATE_PROJECT)
        res = self._request(create_project_url, "post", project_data)
        return res.json()

    def delete_project(self, uuid: int, query_string: str = None) -> bool:
        url = urljoin(self.api_url, self.URL_GET_PROJECT.format(uuid))
        if query_string:
            url = f"{url}?{query_string}"
        res = self._request(url, "delete")
        return res.ok

    def update_project(self, data: dict, query_string: str = None) -> dict:
        url = urljoin(self.api_url, self.URL_GET_PROJECT.format(data["id"]))
        if query_string:
            url = f"{url}?{query_string}"
        res = self._request(url, "put", data)
        return res.json()

    def attach_files(
            self,
            project_id: int,
            folder_id: int,
            team_id: int,
            files: List[Dict],
            annotation_status_code,
            upload_state_code,
            meta,
    ):
        data = {
            "project_id": project_id,
            "folder_id": folder_id,
            "team_id": team_id,
            "images": files,
            "annotation_status": annotation_status_code,
            "upload_state": upload_state_code,
            "meta": meta,
        }
        create_image_url = urljoin(self.api_url, self.URL_CREATE_IMAGE)
        response = self._request(create_image_url, "post", data)
        return response.json()

    def get_folder(self, query_string: str):
        get_folder_url = urljoin(self.api_url, self.URL_GET_FOLDER_BY_NAME)
        if query_string:
            get_folder_url = f"{get_folder_url}?{query_string}"
        response = self._request(get_folder_url, "get")
        if response.ok:
            return response.json()

    def get_folders(self, query_string: str = None, params: dict = None):
        get_folder_url = urljoin(self.api_url, self.URL_FOLDERS_IMAGES)
        if query_string:
            get_folder_url = f"{get_folder_url}?{query_string}"
        return self._get_all_pages(get_folder_url, params=params, key_field="folders")

    def delete_folders(self, project_id: int, team_id: int, folder_ids: List[int]):
        delete_folders_url = urljoin(self.api_url, self.URL_DELETE_FOLDERS)
        params = {"team_id": team_id, "project_id": project_id}
        response = self._request(
            delete_folders_url, "put", params=params, data={"folder_ids": folder_ids}
        )
        return response.ok

    def create_folder(self, project_id: int, team_id: int, folder_name: str):
        create_folder_url = urljoin(self.api_url, self.URL_CREATE_FOLDER)
        data = {"name": folder_name}
        params = {"project_id": project_id, "team_id": team_id}
        res = self._request(create_folder_url, "post", data=data, params=params)
        return res.json()

    def update_folder(self, project_id: int, team_id: int, folder_data: dict):
        update_folder_url = urljoin(
            self.api_url, self.URL_UPDATE_FOLDER.format(folder_data["id"])
        )
        params = {"project_id": project_id, "team_id": team_id}
        res = self._request(update_folder_url, "put", data=folder_data, params=params)
        if res.ok:
            return res.json()

    def get_project_settings(self, project_id: int, team_id: int):
        get_settings_url = urljoin(
            self.api_url, self.URL_PROJECT_SETTINGS.format(project_id)
        )
        res = self._request(get_settings_url, "get", params={"team_id": team_id})
        return res.json()

    def set_project_settings(self, project_id: int, team_id: int, data: List):
        set_project_settings_url = urljoin(
            self.api_url, self.URL_PROJECT_SETTINGS.format(project_id)
        )
        res = self._request(
            set_project_settings_url,
            "put",
            data={"settings": data},
            params={"team_id": team_id},
        )
        return res.json()

    def get_annotation_classes(
            self, project_id: int, team_id: int, query_string: str = None
    ):
        get_annotation_classes_url = urljoin(self.api_url, self.URL_ANNOTATION_CLASSES)
        if query_string:
            get_annotation_classes_url = f"{get_annotation_classes_url}?{query_string}"
        params = {"project_id": project_id, "team_id": team_id}
        return self._get_all_pages(get_annotation_classes_url, params=params)

    def set_annotation_classes(self, project_id: int, team_id: int, data: List):
        set_annotation_class_url = urljoin(self.api_url, self.URL_ANNOTATION_CLASSES)
        params = {
            "team_id": team_id,
            "project_id": project_id,
        }
        res = self._request(
            set_annotation_class_url, "post", params=params, data={"classes": data}
        )
        return res.json()

    def get_project_workflows(self, project_id: int, team_id: int):
        get_project_workflow_url = urljoin(
            self.api_url, self.URL_PROJECT_WORKFLOW.format(project_id)
        )
        return self._get_all_pages(
            get_project_workflow_url, params={"team_id": team_id}
        )

    def set_project_workflow(self, project_id: int, team_id: int, data: Dict):
        set_project_workflow_url = urljoin(
            self.api_url, self.URL_PROJECT_WORKFLOW.format(project_id)
        )
        res = self._request(
            set_project_workflow_url,
            "post",
            data={"steps": [data]},
            params={"team_id": team_id},
        )
        return res.json()

    def set_project_workflow_bulk(self, project_id: int, team_id: int, steps: list):
        set_project_workflow_url = urljoin(
            self.api_url, self.URL_PROJECT_WORKFLOW.format(project_id)
        )
        res = self._request(
            set_project_workflow_url,
            "post",
            data={"steps": steps},
            params={"team_id": team_id},
        )
        return res.json()

    def share_project_bulk(self, project_id: int, team_id: int, users: list):
        share_project_url = urljoin(
            self.api_url, self.URL_SHARE_PROJECT_BULK.format(project_id)
        )
        res = self._request(
            share_project_url,
            "post",
            data={"users": users},
            params={"team_id": team_id},
        )
        return res.json()

    def search_team_contributors(self, team_id: int, query_string: str = None):

        list_users_url = urljoin(self.api_url, self.URL_USERS)
        if query_string:
            list_users_url = f"{list_users_url}?{query_string}"
        params = {"team_id": team_id}
        return self._get_all_pages(list_users_url, params=params)

    def un_share_project(self, project_id: int, team_id: int, user_id: int):
        users_url = urljoin(self.api_url, self.URL_SHARE_PROJECT.format(project_id))

        res = self._request(
            users_url, "delete", data={"user_id": user_id}, params={"team_id": team_id}
        )
        return res.ok

    def get_images(self, query_string: str = None):
        url = urljoin(self.api_url, self.URL_FOLDERS_IMAGES)
        if query_string:
            url = f"{url}?{query_string}"
        pages = self._get_all_pages(url, key_field="images")
        return [image for image in pages]

    def list_images(self, query_string):
        url = urljoin(self.api_url, self.URL_GET_IMAGES)
        if query_string:
            url = f"{url}?{query_string}"
        return self._get_all_pages(url)

    def list_items(self, query_string) -> List[dict]:
        url = urljoin(self.api_url, self.URL_GET_ITEMS)
        if query_string:
            url = f"{url}?{query_string}"
        return self._get_all_pages(url)

    def prepare_export(
            self,
            project_id: int,
            team_id: int,
            folders: List[str],
            annotation_statuses: Iterable[str],
            include_fuse: bool,
            only_pinned: bool,
    ):
        prepare_export_url = urljoin(self.api_url, self.URL_PREPARE_EXPORT)

        annotation_statuses = ",".join(
            [str(constance.AnnotationStatus.get_value(i)) for i in annotation_statuses]
        )

        data = {
            "include": annotation_statuses,
            "fuse": int(include_fuse),
            "is_pinned": int(only_pinned),
            "coco": 0,
            "time": datetime.datetime.now().strftime("%b %d %Y %H:%M"),
        }
        if folders:
            data["folder_names"] = folders

        res = self._request(
            prepare_export_url,
            "post",
            data=data,
            params={"project_id": project_id, "team_id": team_id},
        )
        return res.json()

    def get_team(self, team_id: int):
        get_team_url = urljoin(self.api_url, f"{self.URL_TEAM}/{team_id}")
        res = self._request(get_team_url, "get")
        return res.json()

    def delete_team_invitation(self, team_id: int, token: str, email: str) -> bool:
        invite_contributor_url = urljoin(
            self.api_url, self.URL_INVITE_CONTRIBUTOR.format(team_id)
        )
        res = self._request(
            invite_contributor_url, "delete", data={"token": token, "e_mail": email}
        )
        return res.ok

    def invite_contributors(
            self, team_id: int, team_role: int, emails: list
    ) -> Tuple[List[str], List[str]]:
        invite_contributors_url = urljoin(
            self.api_url, self.URL_INVITE_CONTRIBUTORS.format(team_id)
        )
        res = self._request(
            invite_contributors_url,
            "post",
            data=dict(emails=emails, team_role=team_role),
        ).json()
        return res["success"]["emails"], res["failed"]["emails"]

    def update_image(self, image_id: int, team_id: int, project_id: int, data: dict):
        update_image_url = urljoin(self.api_url, self.URL_GET_IMAGE.format(image_id))

        res = self._request(
            update_image_url,
            "put",
            data=data,
            params={"team_id": team_id, "project_id": project_id},
        )
        return res.ok

    def copy_items_between_folders_transaction(
            self,
            team_id: int,
            project_id: int,
            from_folder_id: int,
            to_folder_id: int,
            items: List[str],
            include_annotations: bool = False,
            include_pin: bool = False,
    ) -> int:
        """
        Returns poll id.
        """
        copy_images_url = urljoin(self.api_url, self.URL_COPY_IMAGES_FROM_FOLDER)
        res = self._request(
            copy_images_url,
            "post",
            params={"team_id": team_id, "project_id": project_id},
            data={
                "is_folder_copy": False,
                "image_names": items,
                "destination_folder_id": to_folder_id,
                "source_folder_id": from_folder_id,
                "include_annotations": include_annotations,
                "keep_pin_status": include_pin,
            },
        )
        if res.ok:
            return res.json()["poll_id"]

    def move_images_between_folders(
            self,
            team_id: int,
            project_id: int,
            from_folder_id: int,
            to_folder_id: int,
            images: List[str],
    ) -> List[str]:
        move_images_url = urljoin(self.api_url, self.URL_MOVE_IMAGES_FROM_FOLDER)
        res = self._request(
            move_images_url,
            "post",
            params={"team_id": team_id, "project_id": project_id},
            data={
                "image_names": images,
                "destination_folder_id": to_folder_id,
                "source_folder_id": from_folder_id,
            },
        )
        if res.ok:
            return res.json()["done"]
        return []

    def get_progress(
            self, project_id: int, team_id: int, poll_id: int
    ) -> Tuple[int, int]:
        get_progress_url = urljoin(self.api_url, self.URL_GET_COPY_PROGRESS)

        res = self._request(
            get_progress_url,
            "get",
            params={"team_id": team_id, "project_id": project_id, "poll_id": poll_id},
        ).json()
        return res["done"], res["skipped"]

    def await_progress(self, project_id: int, team_id: int, poll_id: int, items_count):
        try:
            await_time = items_count * 0.3
            timeout_start = time.time()
            while time.time() < timeout_start + await_time:
                done_count, skipped = self.get_progress(project_id, team_id, poll_id)
                if done_count + skipped == items_count:
                    break
                time.sleep(4)
        except (AppException, Exception) as e:
            raise BackendError(e)

    def get_duplicated_images(
            self, project_id: int, team_id: int, folder_id: int, images: List[str]
    ) -> List[str]:
        get_duplications_url = urljoin(self.api_url, self.URL_BULK_GET_IMAGES)

        res = self._request(
            get_duplications_url,
            "post",
            data={
                "project_id": project_id,
                "team_id": team_id,
                "folder_id": folder_id,
                "names": images,
            },
        )
        return res.json()

    def delete_image(self, image_id, team_id: int, project_id: int):
        delete_image_url = urljoin(self.api_url, self.URL_GET_IMAGE.format(image_id))

        res = self._request(
            delete_image_url,
            "delete",
            params={"team_id": team_id, "project_id": project_id},
        )
        return res.ok

    def set_images_statuses_bulk(
            self,
            image_names: list,
            team_id: int,
            project_id: int,
            folder_id: int,
            annotation_status: int,
    ):
        set_images_statuses_bulk_url = urljoin(
            self.api_url, self.URL_SET_IMAGES_STATUSES_BULK
        )

        res = self._request(
            set_images_statuses_bulk_url,
            "put",
            params={"team_id": team_id, "project_id": project_id},
            data={
                "folder_id": folder_id,
                "annotation_status": annotation_status,
                "image_names": image_names,
            },
        )
        return res.ok

    def get_bulk_images(
            self, project_id: int, team_id: int, folder_id: int, images: List[str]
    ) -> List[dict]:
        bulk_get_images_url = urljoin(self.api_url, self.URL_BULK_GET_IMAGES)
        res = self._request(
            bulk_get_images_url,
            "post",
            data={
                "project_id": project_id,
                "team_id": team_id,
                "folder_id": folder_id,
                "names": images,
            },
        )
        return res.json()

    def delete_images(self, project_id: int, team_id: int, image_ids: List[int]):
        delete_images_url = urljoin(self.api_url, self.URL_DELETE_FOLDERS)
        res = self._request(
            delete_images_url,
            "put",
            params={"team_id": team_id, "project_id": project_id},
            data={"image_ids": image_ids},
        )
        return res.json()

    def assign_images(
            self,
            team_id: int,
            project_id: int,
            folder_name: str,
            user: str,
            image_names: list,
    ):
        assign_images_url = urljoin(self.api_url, self.URL_ASSIGN_IMAGES)
        res = self._request(
            assign_images_url,
            "put",
            params={"team_id": team_id, "project_id": project_id},
            data={
                "image_names": image_names,
                "assign_user_id": user,
                "folder_name": folder_name,
            },
        )
        return res.ok

    def un_assign_images(
            self, team_id: int, project_id: int, folder_name: str, image_names: List[str],
    ):
        un_assign_images_url = urljoin(self.api_url, self.URL_ASSIGN_IMAGES)
        res = self._request(
            un_assign_images_url,
            "put",
            params={"team_id": team_id, "project_id": project_id},
            data={
                "image_names": image_names,
                "remove_user_ids": ["all"],
                "folder_name": folder_name,
            },
        )
        return res.ok

    def un_assign_folder(
            self, team_id: int, project_id: int, folder_name: str,
    ):
        un_assign_folder_url = urljoin(self.api_url, self.URL_ASSIGN_FOLDER)
        res = self._request(
            un_assign_folder_url,
            "post",
            params={"team_id": team_id, "project_id": project_id},
            data={"folder_name": folder_name, "remove_user_ids": ["all"]},
        )
        return res.ok

    def assign_folder(
            self, team_id: int, project_id: int, folder_name: str, users: list
    ):
        assign_folder_url = urljoin(self.api_url, self.URL_ASSIGN_FOLDER)
        res = self._request(
            assign_folder_url,
            "post",
            params={"team_id": team_id, "project_id": project_id},
            data={"folder_name": folder_name, "assign_user_ids": users},
        )
        return res.ok

    def get_exports(self, team_id: int, project_id: int):
        exports_url = urljoin(self.api_url, self.URL_GET_EXPORTS)
        res = self._request(
            exports_url, "get", params={"team_id": team_id, "project_id": project_id}
        )
        return res.json()

    def get_export(self, team_id: int, project_id: int, export_id: int):
        exports_url = urljoin(self.api_url, self.URL_GET_EXPORT.format(export_id))
        res = self._request(
            exports_url, "get", params={"team_id": team_id, "project_id": project_id}
        )
        return res.json()

    def upload_form_s3(
            self,
            project_id: int,
            team_id: int,
            access_key: str,
            secret_key: str,
            bucket_name: str,
            from_folder_name: str,
            to_folder_id: int,
    ):
        upload_from_s3_url = urljoin(
            self.api_url, self.URL_S3_ACCESS_POINT.format(project_id)
        )
        response = self._request(
            upload_from_s3_url,
            "post",
            params={"team_id": team_id},
            data={
                "accessKeyID": access_key,
                "secretAccessKey": secret_key,
                "bucketName": bucket_name,
                "folderName": from_folder_name,
                "folder_id": to_folder_id,
            },
        )
        return response

    def get_upload_status(self, project_id: int, team_id: int, folder_id: int):
        get_upload_status_url = urljoin(
            self.api_url, self.URL_S3_UPLOAD_STATUS.format(project_id)
        )
        res = self._request(
            get_upload_status_url,
            "get",
            params={"team_id": team_id, "folder_id": folder_id},
        )
        return res.json().get("progress")

    def get_project_images_count(self, team_id: int, project_id: int):
        get_images_count_url = urljoin(self.api_url, self.URL_FOLDERS_IMAGES)
        res = self._request(
            get_images_count_url,
            "get",
            params={"team_id": team_id, "project_id": project_id},
        )
        return res.json()

    def delete_annotation_class(
            self, team_id: int, project_id: int, annotation_class_id: int
    ):
        delete_image_url = urljoin(
            self.api_url, self.URL_GET_CLASS.format(annotation_class_id)
        )
        res = self._request(
            delete_image_url,
            "delete",
            params={"team_id": team_id, "project_id": project_id},
        )
        return res.json()

    def set_project_workflow_attributes_bulk(
            self, project_id: int, team_id: int, attributes: list
    ):
        set_project_workflow_attribute_url = urljoin(
            self.api_url, self.URL_PROJECT_WORKFLOW_ATTRIBUTE.format(project_id)
        )
        res = self._request(
            set_project_workflow_attribute_url,
            "post",
            data={"data": attributes},
            params={"team_id": team_id},
        )
        return res.json()

    def get_annotation_upload_data(
            self, project_id: int, team_id: int, image_ids: List[int], folder_id: int
    ):
        get_annotation_upload_data_url = urljoin(
            self.api_url, self.URL_ANNOTATION_UPLOAD_PATH_TOKEN
        )
        response = self._request(
            get_annotation_upload_data_url,
            "post",
            data={
                "project_id": project_id,
                "team_id": team_id,
                "ids": image_ids,
                "folder_id": folder_id,
            },
            content_type=UploadAnnotationAuthData,
        )
        return response

    def get_pre_annotation_upload_data(
            self, project_id: int, team_id: int, image_ids: List[int], folder_id: int
    ):
        get_annotation_upload_data_url = urljoin(
            self.api_url, self.URL_PRE_ANNOTATION_UPLOAD_PATH_TOKEN
        )
        response = self._request(
            get_annotation_upload_data_url,
            "post",
            data={
                "project_id": project_id,
                "team_id": team_id,
                "ids": image_ids,
                "folder_id": folder_id,
            },
            content_type=UploadAnnotationAuthData,
        )
        return response

    def get_templates(self, team_id: int):
        get_templates_url = urljoin(self.api_url, self.URL_GET_TEMPLATES)
        response = self._request(get_templates_url, "get", params={"team_id": team_id})
        return response.json()

    def start_model_training(self, team_id: int, hyper_parameters: dict) -> dict:
        start_training_url = urljoin(self.api_url, self.URL_MODELS)

        res = self._request(
            start_training_url,
            "post",
            params={"team_id": team_id},
            data=hyper_parameters,
        )
        return res.json()

    def get_model_metrics(self, team_id: int, model_id: int) -> dict:
        get_metrics_url = urljoin(
            self.api_url, self.URL_GET_MODEL_METRICS.format(model_id)
        )
        res = self._request(get_metrics_url, "get", params={"team_id": team_id})
        return res.json()

    def get_models(
            self, name: str, team_id: int, project_id: int, model_type: str
    ) -> List:
        search_model_url = urljoin(self.api_url, self.URL_MODELS)
        res = self._request(
            search_model_url,
            "get",
            params={"team_id": team_id, "project_id": project_id, "name": name},
        )
        return res.json()

    def search_models(self, query_string: str):
        search_model_url = urljoin(self.api_url, self.URL_MODELS)
        if query_string:
            search_model_url = f"{search_model_url}?{query_string}"
        # response = self._request(search_model_url, "get",)
        return self._get_all_pages(search_model_url)

    def bulk_get_folders(self, team_id: int, project_ids: List[int]):
        get_folders_url = urljoin(self.api_url, self.URL_BULK_GET_FOLDERS)
        res = self._request(
            get_folders_url,
            "put",
            params={"team_id": team_id, "completedImagesCount": True},
            data={"project_ids": project_ids},
        )
        return res.json()

    def update_model(self, team_id: int, model_id: int, data: dict):
        update_model_url = urljoin(self.api_url, f"{self.URL_MODELS}/{model_id}")
        res = self._request(
            update_model_url, "put", data=data, params={"team_id": team_id}
        )
        return res.json()

    def delete_model(self, team_id: int, model_id: int):
        delete_model_url = urljoin(self.api_url, f"{self.URL_MODEL}/{model_id}")
        res = self._request(delete_model_url, "delete", params={"team_id": team_id})
        return res.ok

    def get_ml_model_download_tokens(self, team_id: int, model_id: int):
        get_token_url = urljoin(
            self.api_url, self.URL_GET_ML_MODEL_DOWNLOAD_TOKEN.format(model_id)
        )
        return self._request(
            get_token_url,
            "get",
            params={"team_id": team_id},
            content_type=DownloadMLModelAuthData,
        )

    def run_prediction(
            self, team_id: int, project_id: int, ml_model_id: int, image_ids: list
    ):
        prediction_url = urljoin(self.api_url, self.URL_PREDICTION)
        res = self._request(
            prediction_url,
            "post",
            data={
                "team_id": team_id,
                "project_id": project_id,
                "ml_model_id": ml_model_id,
                "image_ids": image_ids,
            },
        )
        return res

    def delete_image_annotations(
            self,
            team_id: int,
            project_id: int,
            folder_id: int = None,
            image_names: List[str] = None,
    ) -> dict:
        delete_annotations_url = urljoin(self.api_url, self.URL_DELETE_ANNOTATIONS)
        params = {"team_id": team_id, "project_id": project_id}
        data = {}
        if folder_id:
            params["folder_id"] = folder_id
        if image_names:
            data["image_names"] = image_names
        response = self._request(
            delete_annotations_url, "post", params=params, data=data
        )
        if response.ok:
            return response.json()

    def get_annotations_delete_progress(
            self, team_id: int, project_id: int, poll_id: int
    ):
        get_progress_url = urljoin(self.api_url, self.URL_DELETE_ANNOTATIONS_PROGRESS)

        response = self._request(
            get_progress_url,
            "get",
            params={"team_id": team_id, "project_id": project_id, "poll_id": poll_id},
        )
        return response.json()

    def get_limitations(
            self, team_id: int, project_id: int, folder_id: int = None
    ) -> ServiceResponse:
        get_limits_url = urljoin(self.api_url, self.URL_GET_LIMITS.format(project_id))
        return self._request(
            get_limits_url,
            "get",
            params={"team_id": team_id, "folder_id": folder_id},
            content_type=UserLimits,
        )

    def get_annotations(
            self,
            project_id: int,
            team_id: int,
            folder_id: int,
            items: List[str],
            reporter: Reporter,
    ) -> List[dict]:
        import nest_asyncio

        nest_asyncio.apply()

        query_params = {
            "team_id": team_id,
            "project_id": project_id,
        }
        if folder_id:
            query_params["folder_id"] = folder_id

        handler = StreamedAnnotations(self.default_headers, reporter)
        loop = asyncio.new_event_loop()

        return loop.run_until_complete(
            handler.get_data(
                url=urljoin(self.assets_provider_url, self.URL_GET_ANNOTATIONS),
                data=items,
                params=query_params,
                chunk_size=self.DEFAULT_CHUNK_SIZE,
                map_function=lambda x: {"image_names": x},
            )
        )

    def get_integrations(self, team_id: int) -> List[dict]:
        get_integrations_url = urljoin(
            self.api_url, self.URL_GET_INTEGRATIONS.format(team_id)
        )

        response = self._request(
            get_integrations_url, "get", params={"team_id": team_id}
        )
        if response.ok:
            return response.json().get("integrations", [])
        return []

    def attach_integrations(
            self,
            team_id: int,
            project_id: int,
            integration_id: int,
            folder_id: int,
            folder_name: str = None,
    ) -> bool:
        attach_integrations_url = urljoin(
            self.api_url, self.URL_ATTACH_INTEGRATIONS.format(team_id)
        )
        data = {
            "team_id": team_id,
            "project_id": project_id,
            "folder_id": folder_id,
            "integration_id": integration_id,
        }
        if folder_name:
            data["customer_folder_name"] = folder_name
        response = self._request(attach_integrations_url, "post", data=data)
        return response.ok

    def saqul_query(
            self, team_id: int, project_id: int, query: str, folder_id: int
    ) -> ServiceResponse:
        CHUNK_SIZE = 50
        query_url = urljoin(self.api_url, self.URL_SAQUL_QUERY)
        params = {
            "team_id": team_id,
            "project_id": project_id,
        }
        if folder_id:
            params["folder_id"] = folder_id
        data = {"query": query, "image_index": 0}
        items = []
        for _ in range(self.MAX_ITEMS_COUNT):
            response = self._request(query_url, "post", params=params, data=data)
            if response.ok:
                response_items = response.json()
                items.extend(response_items)
                if len(response_items) < CHUNK_SIZE:
                    service_response = ServiceResponse(response)
                    service_response.data = items
                    return service_response
                data["image_index"] += CHUNK_SIZE
            return ServiceResponse(response)

    def validate_saqul_query(self, team_id: int, project_id: int, query: str) -> dict:
        validate_query_url = urljoin(self.api_url, self.URL_VALIDATE_SAQUL_QUERY)
        params = {
            "team_id": team_id,
            "project_id": project_id,
        }
        data = {
            "query": query,
        }
        return self._request(
            validate_query_url, "post", params=params, data=data
        ).json()



from lib import get_default_controller
from lib.infrastructure.repositories import ConfigRepository


class BaseInterfaceFacade:
    def __init__(self):
        self._config_path = None
        self._controller = get_default_controller()

    @property
    def controller(self):
        if not ConfigRepository().get_one("token"):
            raise Exception("Config does not exists!")
        return self._controller



"""
COCO to SA conversion method
"""
import json
import threading
from pathlib import Path

import cv2
import numpy as np
from superannotate.logger import get_default_logger

from ....common import blue_color_generator
from ....common import hex_to_rgb
from ....common import id2rgb
from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_pixel_instance
from ..sa_json_helper import _create_sa_json
from .coco_api import _maskfrRLE
from .coco_api import decode

logger = get_default_logger()


def annot_to_bitmask(annot):
    if isinstance(annot["counts"], list):
        bitmask = _maskfrRLE(annot)
    elif isinstance(annot["counts"], str):
        bitmask = decode(annot)

    return bitmask


def coco_panoptic_segmentation_to_sa_pixel(coco_path, output_dir):
    coco_json = json.load(open(coco_path))
    hex_colors = blue_color_generator(len(coco_json["categories"]))

    cat_id_to_cat = {}
    for cat in coco_json["categories"]:
        cat_id_to_cat[cat["id"]] = cat["name"]

    img_id_to_shape = {}
    for img in coco_json["images"]:
        img_id_to_shape[str(img["id"])] = {
            "height": img["height"],
            "width": img["width"],
        }

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(coco_json["annotations"]),
            images_converted,
            images_not_converted,
            finish_event,
        ),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for annot in coco_json["annotations"]:
        annot_name = Path(annot["file_name"]).stem
        img_cv = cv2.imread(str(output_dir / ("%s.png" % annot_name)))
        if img_cv is None:
            images_not_converted.append(annot["file_name"])
            logger.warning(
                "'%s' file dosen't exist!", output_dir / ("%s.png" % annot_name)
            )
            continue

        img = cv2.cvtColor(img_cv, cv2.COLOR_BGR2RGB)
        H, W, C = img.shape
        img = img.reshape((H * W, C))
        segments = annot["segments_info"]
        hex_colors = blue_color_generator(len(segments))

        sa_instances = []
        for i, seg in enumerate(segments):
            img[np.all(img == id2rgb(seg["id"]), axis=1)] = hex_to_rgb(hex_colors[i])
            parts = [{"color": hex_colors[i]}]
            sa_obj = _create_pixel_instance(
                parts, [], cat_id_to_cat[seg["category_id"]]
            )
            sa_instances.append(sa_obj)

        img = cv2.cvtColor(img.reshape((H, W, C)), cv2.COLOR_RGB2BGR)
        cv2.imwrite(str(output_dir / ("%s___save.png" % annot["file_name"])), img)

        images_converted.append(annot["file_name"])
        file_name = "%s___pixel.json" % annot["file_name"]
        sa_metadata = {
            "name": annot_name,
            "width": img_id_to_shape[str(annot["image_id"])]["width"],
            "height": img_id_to_shape[str(annot["image_id"])]["height"],
        }
        json_template = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, json_template)
        (output_dir / ("%s.png" % annot_name)).unlink()
    finish_event.set()
    tqdm_thread.join()


def coco_instance_segmentation_to_sa_pixel(coco_path, output_dir):
    coco_json = json.load(open(coco_path))
    cat_id_to_cat = {}
    for cat in coco_json["categories"]:
        cat_id_to_cat[cat["id"]] = cat

    images_dict = {}
    for img in coco_json["images"]:
        images_dict[int(img["id"])] = {
            "shape": (img["height"], img["width"], 4),
            "file_name": img["file_name"],
            "annotations": [],
        }

    for annot in coco_json["annotations"]:
        if int(annot["image_id"]) not in images_dict:
            continue
        images_dict[annot["image_id"]]["annotations"].append(annot)

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(images_dict.items()),
            images_converted,
            images_not_converted,
            finish_event,
        ),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for id_, annotations in images_dict.items():
        file_name = "%s___pixel.json" % annotations["file_name"]
        hexcolors = blue_color_generator(len(annotations["annotations"]))
        mask = np.zeros(annotations["shape"])
        H, W, _ = mask.shape

        sa_instances = []
        for i, annot in enumerate(annotations["annotations"]):
            hexcolor = hexcolors[i]
            color = hex_to_rgb(hexcolor)
            if isinstance(annot["segmentation"], dict):
                bitmask = annot_to_bitmask(annot["segmentation"])
                mask[bitmask == 1] = list(color)[::-1] + [255]
            else:
                for segment in annot["segmentation"]:
                    bitmask = np.zeros((H, W)).astype(np.uint8)
                    pts = np.array(
                        [
                            segment[2 * i : 2 * (i + 1)]
                            for i in range(len(segment) // 2)
                        ],
                        dtype=np.int32,
                    )
                    cv2.fillPoly(bitmask, [pts], 1)
                    mask[bitmask == 1] = list(color)[::-1] + [255]

            parts = [{"color": hexcolor}]
            sa_obj = _create_pixel_instance(
                parts, [], cat_id_to_cat[annot["category_id"]]["name"]
            )
            sa_instances.append(sa_obj)

        sa_metadata = {
            "name": images_dict[id_]["file_name"],
            "width": images_dict[id_]["shape"][1],
            "height": images_dict[id_]["shape"][0],
        }
        images_converted.append(annotations["file_name"])
        json_template = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, json_template)
        cv2.imwrite(
            str(output_dir / ("%s___save.png" % annotations["file_name"])), mask
        )
    finish_event.set()
    tqdm_thread.join()



from collections import defaultdict
from typing import Union

import tqdm
from superannotate.logger import get_default_logger


class Reporter:
    def __init__(
        self,
        log_info: bool = True,
        log_warning: bool = True,
        disable_progress_bar: bool = False,
        log_debug: bool = True,
    ):
        self.logger = get_default_logger()
        self._log_info = log_info
        self._log_warning = log_warning
        self._log_debug = log_debug
        self._disable_progress_bar = disable_progress_bar
        self.info_messages = []
        self.warning_messages = []
        self.debug_messages = []
        self.custom_messages = defaultdict(set)
        self.progress_bar = None

    def disable_warnings(self):
        self._log_warning = False

    def disable_info(self):
        self._log_info = False

    def enable_warnings(self):
        self._log_warning = True

    def enable_info(self):
        self._log_info = True

    def log_info(self, value: str):
        if self._log_info:
            self.logger.info(value)
        self.info_messages.append(value)

    def log_warning(self, value: str):
        if self._log_warning:
            self.logger.warning(value)
        self.warning_messages.append(value)

    def log_debug(self, value: str):
        if self._log_debug:
            self.logger.debug(value)
        self.debug_messages.append(value)

    def start_progress(
        self,
        iterations: Union[int, range],
        description: str = "Processing",
        disable=False,
    ):
        self.progress_bar = self.get_progress_bar(iterations, description, disable)

    @staticmethod
    def get_progress_bar(
        iterations: Union[int, range], description: str = "Processing", disable=False
    ):
        if isinstance(iterations, range):
            return tqdm.tqdm(iterations, desc=description, disable=disable)
        else:
            return tqdm.tqdm(total=iterations, desc=description, disable=disable)

    def finish_progress(self):
        self.progress_bar.close()

    def update_progress(self, value: int = 1):
        if self.progress_bar:
            self.progress_bar.update(value)

    def generate_report(self) -> str:
        report = ""
        if self.info_messages:
            report += "\n".join(self.info_messages)
        if self.warning_messages:
            report += "\n".join(self.warning_messages)
        return report

    def store_message(self, key: str, value: str):
        self.custom_messages[key].add(value)

    @property
    def messages(self):
        for key, values in self.custom_messages.items():
            yield f"{key} [{', '.join(values)}]"


class Progress:
    def __init__(self, iterations: Union[int, range], description: str = "Processing"):
        self._iterations = iterations
        self._description = description
        self._progress_bar = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self._progress_bar:
            self._progress_bar.close()

    def update(self, value=1):
        if not self._progress_bar:
            self._progress_bar = Reporter.get_progress_bar(
                self._iterations, self._description
            )
        self._progress_bar.update(value)



from lib.core.entities.base import TimedBaseModel
from lib.core.enums import IntegrationTypeEnum
from pydantic import Extra
from pydantic import Field


class IntegrationEntity(TimedBaseModel):
    id: int = None
    user_id: str = None
    name: str
    type: IntegrationTypeEnum = Field(None, alias="source")
    root: str = Field(None, alias="bucket_name")

    class Config:
        extra = Extra.ignore



import itertools
import math
from collections import defaultdict
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from pydantic import BaseModel


class Annotation(BaseModel):
    instanceId: int
    type: str
    className: str
    points: Optional[Dict]
    attributes: Optional[List[Any]] = []
    keyframe: bool = False


class FrameAnnotation(BaseModel):
    frame: int
    annotations: List[Annotation] = []


class VideoFrameGenerator:
    def __init__(self, annotation_data: dict, fps: int):
        self.id_generator = iter(itertools.count(0))
        self._annotation_data = annotation_data
        self.duration = annotation_data["metadata"]["duration"] / (1000 * 1000)
        self.fps = fps
        self.ratio = 1000 * 1000 / fps
        self._frame_id = 1
        self.frames_count = int(self.duration * fps)
        self.annotations: dict = {}
        self._mapping = {}
        self._process()

    def get_frame(self, frame_no: int):
        try:
            return self.annotations[frame_no]
        except KeyError:
            self.annotations[frame_no] = FrameAnnotation(frame=frame_no)
            return self.annotations[frame_no]

    def interpolate_annotations(
            self,
            class_name: str,
            from_frame: int,
            to_frame: int,
            data: dict,
            instance_id: int,
            steps: dict = None,
            annotation_type: str = "bbox",
    ) -> dict:
        annotations = {}
        for idx, frame_idx in enumerate(range(from_frame + 1, to_frame), 1):
            points = None
            if annotation_type == "bbox" and data.get("points") and steps:
                points = {
                    "x1": round(data["points"]["x1"] + steps["x1"] * idx, 2),
                    "y1": round(data["points"]["y1"] + steps["y1"] * idx, 2),
                    "x2": round(data["points"]["x2"] + steps["x2"] * idx, 2),
                    "y2": round(data["points"]["y2"] + steps["y2"] * idx, 2),
                }
            annotations[frame_idx] = Annotation(
                instanceId=instance_id,
                type=annotation_type,
                className=class_name,
                points=points,
                attributes=data["attributes"],
                keyframe=False,
            )
        return annotations

    def _add_annotation(self, frame_no: int, annotation: Annotation):

        frame = self.get_frame(frame_no)
        frame.annotations.append(annotation)

    @staticmethod
    def pairwise(data: list):
        a, b = itertools.tee(data)
        next(b, None)
        return zip(a, b)

    def get_median(self, annotations: List[dict]) -> dict:
        if len(annotations) == 1:
            return annotations[0]
        first_annotations = annotations[:1][0]
        median = (first_annotations["timestamp"] // self.ratio) + (self.ratio / 2)
        median_annotation = first_annotations
        distance = abs(median - first_annotations["timestamp"])
        for annotation in annotations[1:]:
            annotation_distance = abs(median - annotation["timestamp"])
            if annotation_distance < distance:
                distance = annotation_distance
                median_annotation = annotation
        return median_annotation

    @staticmethod
    def merge_first_frame(frames_mapping):
        try:
            if 0 in frames_mapping:
                frames_mapping[0].extend(frames_mapping[1])
                frames_mapping[1] = frames_mapping[0]
                del frames_mapping[0]
        finally:
            return frames_mapping

    def _process(self):
        for instance in self._annotation_data["instances"]:
            instance_id = next(self.id_generator)
            annotation_type = instance["meta"]["type"]
            class_name = instance["meta"]["className"]
            for parameter in instance["parameters"]:
                frames_mapping = defaultdict(list)
                last_frame_no = None
                last_annotation = None
                interpolated_frames = {}
                for timestamp in parameter["timestamps"]:
                    frames_mapping[
                        int(math.ceil(timestamp["timestamp"] / self.ratio))
                    ].append(timestamp)
                frames_mapping = self.merge_first_frame(frames_mapping)
                for from_frame_no, to_frame_no in self.pairwise(sorted(frames_mapping)):
                    last_frame_no = to_frame_no
                    from_frame, to_frame = (
                        frames_mapping[from_frame_no][-1],
                        frames_mapping[to_frame_no][0],
                    )
                    frames_diff = to_frame_no - from_frame_no
                    if frames_diff > 1:
                        steps = None
                        if (
                                annotation_type == "bbox"
                                and from_frame.get("points")
                                and to_frame.get("points")
                        ):
                            steps = {
                                "y1": round(
                                    (
                                            to_frame["points"]["y1"]
                                            - from_frame["points"]["y1"]
                                    )
                                    / frames_diff,
                                    2,
                                ),
                                "x2": round(
                                    (
                                            to_frame["points"]["x2"]
                                            - from_frame["points"]["x2"]
                                    )
                                    / frames_diff,
                                    2,
                                ),
                                "x1": round(
                                    (
                                            to_frame["points"]["x1"]
                                            - from_frame["points"]["x1"]
                                    )
                                    / frames_diff,
                                    2,
                                ),
                                "y2": round(
                                    (
                                            to_frame["points"]["y2"]
                                            - from_frame["points"]["y2"]
                                    )
                                    / frames_diff,
                                    2,
                                ),
                            }
                        interpolated_frames.update(
                            self.interpolate_annotations(
                                class_name=class_name,
                                from_frame=from_frame_no,
                                to_frame=to_frame_no,
                                data=from_frame,
                                instance_id=instance_id,
                                steps=steps,
                                annotation_type=annotation_type,
                            )
                        )
                    start_median_frame = self.get_median(frames_mapping[from_frame_no])
                    end_median_frame = self.get_median(frames_mapping[to_frame_no])
                    interpolated_frames[from_frame_no] = Annotation(
                        instanceId=instance_id,
                        type=annotation_type,
                        className=class_name,
                        points=start_median_frame.get("points"),
                        attributes=start_median_frame["attributes"],
                        keyframe=True,
                    )
                    last_annotation = Annotation(
                        instanceId=instance_id,
                        type=annotation_type,
                        className=class_name,
                        points=end_median_frame.get("points"),
                        attributes=end_median_frame["attributes"],
                        keyframe=True,
                    )
                self._add_annotation(last_frame_no, last_annotation)
                [
                    self._add_annotation(frame_no, annotation)
                    for frame_no, annotation in interpolated_frames.items()
                ]

    def __iter__(self):
        for frame_no in range(1, int(self.frames_count) + 1):
            yield self.get_frame(frame_no).dict()



"""
Sagemaker to SA conversion method
"""
import json
import threading
from pathlib import Path

import cv2
import numpy as np
from superannotate.logger import get_default_logger

from ....common import blue_color_generator
from ....common import hex_to_rgb
from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_pixel_instance
from ..sa_json_helper import _create_sa_json

logger = get_default_logger()


def sagemaker_instance_segmentation_to_sa_pixel(data_path, output_dir):
    img_mapping = {}
    try:
        img_map_file = open(data_path / "output.manifest")
    except Exception:
        raise Exception("'output.manifest' file doesn't exist")

    for line in img_map_file:
        dd = json.loads(line)
        img_mapping[Path(dd["attribute-name-ref"]).name] = Path(dd["source-ref"]).name

    json_list = list(data_path.glob("*.json"))
    classes_ids = {}
    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(json_list), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for json_file in json_list:
        data_json = json.load(open(json_file))
        for annotataion in data_json:
            if "consolidatedAnnotation" not in annotataion.keys():
                logger.warning("Wrong json files")
                raise Exception

            mask_name = Path(
                annotataion["consolidatedAnnotation"]["content"]["attribute-name-ref"]
            ).name

            classes_dict = annotataion["consolidatedAnnotation"]["content"][
                "attribute-name-ref-metadata"
            ]["internal-color-map"]

            try:
                img = cv2.imread(str(data_path / mask_name.replace(":", "_")))
                H, W, _ = img.shape
            except Exception:
                logger.warning("Can't open %s mask", mask_name.replace(":", "_"))
                images_not_converted.append(mask_name.replace(":", "_"))
                continue

            class_contours = {}
            num_of_contours = 0
            for key, _ in classes_dict.items():
                if classes_dict[key]["class-name"] == "BACKGROUND":
                    continue

                if key not in classes_ids.keys():
                    classes_ids[key] = classes_dict[key]["class-name"]

                bitmask = np.zeros((H, W), dtype=np.int8)
                bitmask[
                    np.all(
                        img == list(hex_to_rgb(classes_dict[key]["hex-color"]))[::-1],
                        axis=2,
                    )
                ] = 255

                bitmask = bitmask.astype(np.uint8)
                contours, _ = cv2.findContours(
                    bitmask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
                )
                class_contours[classes_dict[key]["class-name"]] = contours
                num_of_contours += len(contours)

            blue_colors = blue_color_generator(num_of_contours)
            idx = 0
            file_name = "%s___pixel.json" % (img_mapping[mask_name])
            sa_metadata = {"name": img_mapping[mask_name], "width": W, "height": H}
            sa_instances = []
            sa_mask = np.zeros((H, W, 4))
            for name, contours in class_contours.items():
                parts = []
                for contour in contours:
                    bitmask = np.zeros((H, W))
                    contour = contour.flatten().tolist()
                    pts = np.array(
                        [
                            contour[2 * i : 2 * (i + 1)]  # noqa: E203
                            for i in range(len(contour) // 2)
                        ],
                        dtype=np.int32,
                    )
                    cv2.fillPoly(bitmask, [pts], 1)
                    sa_mask[bitmask == 1] = list(hex_to_rgb(blue_colors[idx]))[::-1] + [
                        255
                    ]
                    parts.append({"color": blue_colors[idx]})
                    idx += 1
                    sa_obj = _create_pixel_instance(parts, [], name)
                    sa_instances.append(sa_obj)
            images_converted.append(img_mapping[mask_name])
            sa_json = _create_sa_json(sa_instances, sa_metadata)
            write_to_json(output_dir / file_name, sa_json)
            cv2.imwrite(
                str(output_dir / (img_mapping[mask_name] + "___save.png")), sa_mask
            )
    finish_event.set()
    tqdm_thread.join()
    return classes_ids



"""
VGG to SA conversion method.
"""
import json
import threading

import cv2
from superannotate.logger import get_default_logger

from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_sa_json
from ..sa_json_helper import _create_vector_instance
from .vgg_helper import _create_attribute_list

logger = get_default_logger()


def vgg_to_sa(json_data, task, output_dir):
    images = json.load(open(json_data))
    if task == "object_detection":
        instance_types = ["rect"]
    elif task == "instance_segmentation":
        instance_types = ["polygon"]
    elif task == "vector_annotation":
        instance_types = ["rect", "polygon", "polyline", "point", "ellipse", "circle"]

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(images.items()),
            images_converted,
            images_not_converted,
            finish_event,
        ),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()

    class_id_map = {}
    for _, img in images.items():
        try:
            H, W, _ = cv2.imread(str(output_dir / img["filename"])).shape
        except Exception:
            logger.warning(
                "Can't open %s image. 'height' and 'width' for SA JSON metadata will set to zero",
                img["filename"],
            )
            H = 0
            W = 0

        file_name = "%s___objects.json" % img["filename"]
        sa_metadata = {"name": img["filename"], "width": W, "height": H}
        sa_instances = []
        instances = img["regions"]
        for instance in instances:
            if "type" not in instance["region_attributes"].keys():
                raise KeyError(
                    "'VGG' JSON should contain 'type' key which will \
                    be category name. Please correct JSON file."
                )
            if not isinstance(instance["region_attributes"]["type"], str):
                raise ValueError("Wrong attribute was choosen for 'type' attribute.")

            class_name = instance["region_attributes"]["type"]
            if class_name not in class_id_map.keys():
                class_id_map[class_name] = {"attribute_groups": {}}

            attributes = _create_attribute_list(
                instance["region_attributes"], class_name, class_id_map
            )

            if instance["shape_attributes"]["name"] in instance_types:
                if (
                    instance["shape_attributes"]["name"] == "polygon"
                    or instance["shape_attributes"]["name"] == "polyline"
                ):
                    points = []
                    for x, y in zip(
                        instance["shape_attributes"]["all_points_x"],
                        instance["shape_attributes"]["all_points_y"],
                    ):
                        points.append(x)
                        points.append(y)
                    instance_type = instance["shape_attributes"]["name"]
                elif instance["shape_attributes"]["name"] == "rect":
                    points = (
                        instance["shape_attributes"]["x"],
                        instance["shape_attributes"]["y"],
                        instance["shape_attributes"]["x"]
                        + instance["shape_attributes"]["width"],
                        instance["shape_attributes"]["y"]
                        + instance["shape_attributes"]["height"],
                    )
                    instance_type = "bbox"
                elif instance["shape_attributes"]["name"] == "ellipse":
                    points = (
                        instance["shape_attributes"]["cx"],
                        instance["shape_attributes"]["cy"],
                        instance["shape_attributes"]["rx"],
                        instance["shape_attributes"]["ry"],
                        instance["shape_attributes"]["theta"],
                    )
                    instance_type = "ellipse"
                elif instance["shape_attributes"]["name"] == "circle":
                    points = (
                        instance["shape_attributes"]["cx"],
                        instance["shape_attributes"]["cy"],
                        instance["shape_attributes"]["r"],
                        instance["shape_attributes"]["r"],
                        0,
                    )
                    instance_type = "ellipse"
                elif instance["shape_attributes"]["name"] == "point":
                    points = (
                        instance["shape_attributes"]["cx"],
                        instance["shape_attributes"]["cy"],
                    )
                    instance_type = "point"
                sa_obj = _create_vector_instance(
                    instance_type, points, {}, attributes, class_name
                )
                sa_instances.append(sa_obj)
        images_converted.append(img["filename"])
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)
    finish_event.set()
    tqdm_thread.join()
    return class_id_map



"""
Sagemaker to SA conversion method
"""
import json
import threading
from pathlib import Path

from superannotate.logger import get_default_logger

from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_sa_json
from ..sa_json_helper import _create_vector_instance

logger = get_default_logger()


def sagemaker_object_detection_to_sa_vector(data_path, main_key, output_dir):
    dataset_manifest = []
    try:
        img_map_file = open(data_path / "output.manifest")
    except Exception as e:
        raise Exception("'output.manifest' file doesn't exist")

    for line in img_map_file:
        dataset_manifest.append(json.loads(line))

    json_list = list(data_path.glob("*.json"))
    classes_ids = {}
    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(json_list), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for json_file in json_list:
        data_json = json.load(open(json_file))
        for img in data_json:
            if "consolidatedAnnotation" not in img.keys():
                logger.warning("Wrong json files")
                raise Exception

            manifest = dataset_manifest[int(img["datasetObjectId"])]
            file_name = "%s___objects.json" % Path(manifest["source-ref"]).name

            classes = img["consolidatedAnnotation"]["content"][main_key + "-metadata"][
                "class-map"
            ]
            for key, value in classes.items():
                if key not in classes_ids.keys():
                    classes_ids[key] = value

            image_size = img["consolidatedAnnotation"]["content"][main_key][
                "image_size"
            ][0]

            sa_metadata = {
                "name": Path(manifest["source-ref"]).name,
                "width": image_size["width"],
                "height": image_size["height"],
            }

            annotations = img["consolidatedAnnotation"]["content"][main_key][
                "annotations"
            ]
            sa_instances = []
            for annotation in annotations:
                points = (
                    annotation["left"],
                    annotation["top"],
                    annotation["left"] + annotation["width"],
                    annotation["top"] + annotation["height"],
                )
                sa_obj = _create_vector_instance(
                    "bbox", points, {}, [], classes[str(annotation["class_id"])]
                )
                sa_instances.append(sa_obj.copy())
            images_converted.append(Path(manifest["source-ref"]).name)
            sa_json = _create_sa_json(sa_instances, sa_metadata)
            write_to_json(output_dir / file_name, sa_json)

    finish_event.set()
    tqdm_thread.join()
    return classes_ids



import json

from ....common import write_to_json
from ..baseStrategy import baseStrategy


class SuperviselyStrategy(baseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def to_sa_format(self):
        sa_classes, classes_id_map = self._create_sa_classes()
        json_files = []
        if self.dataset_name != "":
            json_files.append(
                self.export_root / "ds" / "ann" / (self.dataset_name + ".json")
            )
        else:
            files_gen = (self.export_root / "ds" / "ann").glob("*")
            json_files = list(files_gen)

        if (
            self.conversion_algorithm.__name__
            == "supervisely_keypoint_detection_to_sa_vector"
        ):
            meta_json = json.load(open(self.export_root / "meta.json"))
            sa_jsons = self.conversion_algorithm(
                json_files, classes_id_map, meta_json, self.output_dir
            )
        elif (
            self.conversion_algorithm.__name__
            == "supervisely_instance_segmentation_to_sa_pixel"
        ):
            sa_jsons = self.conversion_algorithm(
                json_files, classes_id_map, self.output_dir
            )
        else:
            sa_jsons = self.conversion_algorithm(
                json_files, classes_id_map, self.task, self.output_dir
            )
        (self.output_dir / "classes").mkdir(exist_ok=True)
        write_to_json(self.output_dir / "classes" / "classes.json", sa_classes)

    def _create_sa_classes(self):
        classes_json = json.load(open(self.export_root / "meta.json"))

        attributes = []
        for tag in classes_json["tags"]:
            attributes.append({"name": tag["name"]})

        classes_id_map = {}
        classes_loader = []
        for class_ in classes_json["classes"]:
            group_name = "converted_attributs"
            classes_id_map[class_["title"]] = {
                "attr_group": {"group_name": group_name, "attributes": []}
            }
            for attribute in attributes:
                attribute["groupName"] = group_name
                classes_id_map[class_["title"]]["attr_group"]["attributes"].append(
                    {"name": attribute["name"]}
                )

            if not attributes:
                attribute_groups = []
            else:
                attribute_groups = [
                    {"name": group_name, "is_multiselect": 1, "attributes": attributes}
                ]

            attr_group = {
                "name": class_["title"],
                "color": class_["color"],
                "attribute_groups": attribute_groups,
            }
            classes_loader.append(attr_group)
        return classes_loader, classes_id_map



from packaging.version import parse
from setuptools import find_packages, setup


with open('src/superannotate/version.py') as f:
    version = f.read().rstrip()[15:-1]

requirements_path = "requirements_{}.txt".format('dev' if parse(version).is_prerelease else 'prod')
requirements = []

with open("requirements.txt") as f:
    requirements.extend(f.read().splitlines())

with open(requirements_path) as f:
    requirements.extend(f.read().splitlines())


with open('README.md') as f:
    readme = f.read()

readme = "\n".join(readme.split('\n')[2:])


setup(
    name='superannotate',
    version=version,
    package_dir={"": "src"},
    package_data={"superannotate": ["logging.conf"]},
    packages=find_packages(where="src"),
    description='Python SDK to SuperAnnotate platform',
    license='MIT',
    author='SuperAnnotate AI',
    author_email='vaghinak@superannotate.com',
    url='https://github.com/superannotateai/superannotate-python-sdk',
    long_description=readme,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    setup_requires=['wheel'],
    description_file="README.md",
    entry_points={
        'console_scripts': ['superannotatecli = superannotate.lib.app.bin.superannotate:main']
    },
    python_requires='>=3.7'
)



import logging.config
import os
from logging import Formatter
from logging.handlers import RotatingFileHandler
from os.path import expanduser

from superannotate import constances

default_logger = None

logging.config.dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": False,
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "INFO",
                "formatter": "consoleFormatter",
                "stream": "ext://sys.stdout",
            },
        },
        "formatters": {
            "consoleFormatter": {
                "format": "SA-PYTHON-SDK - %(levelname)s - %(message)s",
            },
        },
        "loggers": {"sa": {"handlers": ["console"], "level": "DEBUG"}},
    }
)


def get_default_logger():
    global default_logger
    if not default_logger:
        default_logger = logging.getLogger("sa")
        try:
            log_file_path = expanduser(constances.LOG_FILE_LOCATION)
            open(log_file_path, "w").close()
            if os.access(log_file_path, os.W_OK):
                file_handler = RotatingFileHandler(
                    expanduser(constances.LOG_FILE_LOCATION),
                    maxBytes=5 * 1024 * 1024,
                    backupCount=5,
                    mode="a",
                )
                formatter = Formatter(
                    "SA-PYTHON-SDK - %(levelname)s - %(asctime)s - %(message)s"
                )
                file_handler.setFormatter(formatter)
                file_handler.setLevel("DEBUG")
                default_logger.addHandler(file_handler)
        except OSError:
            pass
    return default_logger



"""
COCO to SA conversion methods
"""
import json
import threading
from pathlib import Path

import cv2
import numpy as np
from superannotate.logger import get_default_logger

from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_sa_json
from ..sa_json_helper import _create_vector_instance
from .coco_api import _maskfrRLE
from .coco_api import decode

logger = get_default_logger()


def annot_to_polygon(annot):
    if isinstance(annot["counts"], list):
        bitmask = _maskfrRLE(annot)
    elif isinstance(annot["counts"], str):
        bitmask = decode(annot)

    contours, _ = cv2.findContours(
        bitmask.astype(np.uint8), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )
    segments = []
    for contour in contours:
        contour = contour.flatten().tolist()
        if len(contour) > 4:
            segments.append(contour)

    return segments


def save_sa_jsons(coco_json, img_id_to_annot, output_dir):
    images_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(coco_json["images"]), images_converted, [], finish_event),
        daemon=True,
    )
    logger.info("Writting to disk")
    tqdm_thread.start()
    for img in coco_json["images"]:
        if "file_name" in img:
            image_path = Path(img["file_name"]).name
        else:
            image_path = img["coco_url"].split("/")[-1]

        if str(img["id"]) not in img_id_to_annot:
            sa_instances = []
        else:
            sa_instances = img_id_to_annot[str(img["id"])]
        file_name = "%s___objects.json" % image_path

        sa_metadata = {
            "name": image_path,
            "width": img["width"],
            "height": img["height"],
        }
        json_template = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, json_template)
    finish_event.set()
    tqdm_thread.join()


def coco_instance_segmentation_to_sa_vector(coco_path, output_dir):
    coco_json = json.load(open(coco_path))
    cat_id_to_cat = {}
    for cat in coco_json["categories"]:
        cat_id_to_cat[cat["id"]] = cat

    instance_groups = {}
    for annot in coco_json["annotations"]:
        if "id" in annot:
            if annot["id"] in instance_groups:
                instance_groups[annot["id"]] += 1
            else:
                instance_groups[annot["id"]] = 1

    image_id_to_annotations = {}
    annotations_processed = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(coco_json["annotations"]), annotations_processed, [], finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for annot in coco_json["annotations"]:
        if isinstance(annot["segmentation"], dict):
            annot["segmentation"] = annot_to_polygon(annot["segmentation"])

        cat = cat_id_to_cat[annot["category_id"]]
        groupid = 0
        for polygon in annot["segmentation"]:
            if (
                "id" in annot
                and annot["id"] in instance_groups
                and instance_groups[annot["id"]] > 1
            ):
                groupid = annot["id"]

            sa_obj = _create_vector_instance("polygon", polygon, {}, [], cat["name"])
            if groupid != 0:
                sa_obj["groupId"] = groupid

            if str(annot["image_id"]) not in image_id_to_annotations:
                image_id_to_annotations[str(annot["image_id"])] = [sa_obj]
            else:
                image_id_to_annotations[str(annot["image_id"])].append(sa_obj)
        annotations_processed.append(annot)
    finish_event.set()
    tqdm_thread.join()
    save_sa_jsons(coco_json, image_id_to_annotations, output_dir)


def coco_object_detection_to_sa_vector(coco_path, output_dir):
    coco_json = json.load(open(coco_path))
    cat_id_to_cat = {}
    for cat in coco_json["categories"]:
        cat_id_to_cat[cat["id"]] = cat

    image_id_to_annotations = {}
    annotations_processed = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(coco_json["annotations"]), annotations_processed, [], finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for annot in coco_json["annotations"]:
        if isinstance(annot["segmentation"], dict):
            annot["segmentation"] = annot_to_polygon(annot["segmentation"])
        cat = cat_id_to_cat[annot["category_id"]]

        points = (
            annot["bbox"][0],
            annot["bbox"][1],
            annot["bbox"][0] + annot["bbox"][2],
            annot["bbox"][1] + annot["bbox"][3],
        )

        sa_obj = _create_vector_instance("bbox", points, {}, [], cat["name"])

        if str(annot["image_id"]) not in image_id_to_annotations:
            image_id_to_annotations[str(annot["image_id"])] = [sa_obj]
        else:
            image_id_to_annotations[str(annot["image_id"])].append(sa_obj)
        annotations_processed.append(annot)
    finish_event.set()
    tqdm_thread.join()
    save_sa_jsons(coco_json, image_id_to_annotations, output_dir)


def coco_keypoint_detection_to_sa_vector(coco_path, output_dir):
    coco_json = json.load(open(coco_path))

    cat_id_to_cat = {}
    for cat in coco_json["categories"]:
        cat_id_to_cat[cat["id"]] = {
            "name": cat["name"],
            "keypoints": cat["keypoints"],
            "skeleton": cat["skeleton"],
            "supercategory": cat["supercategory"],
        }

    image_id_to_annotations = {}
    annotations_processed = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(coco_json["annotations"]), annotations_processed, [], finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for annot in coco_json["annotations"]:
        if annot["num_keypoints"] > 0:
            sa_points = [
                item
                for index, item in enumerate(annot["keypoints"])
                if (index + 1) % 3 != 0
            ]

            sa_points = [
                (sa_points[i], sa_points[i + 1]) for i in range(0, len(sa_points), 2)
            ]
            if annot["category_id"] in cat_id_to_cat.keys():
                keypoint_names = cat_id_to_cat[annot["category_id"]]["keypoints"]

                bad_points = []
                id_mapping = {}
                index = 1
                points = []
                for point_index, point in enumerate(sa_points):
                    if sa_points[point_index] == (0, 0):
                        bad_points.append(point_index + 1)
                        continue
                    id_mapping[point_index + 1] = index
                    points.append({"id": index, "x": point[0], "y": point[1]})
                    index += 1

                connections = []
                for connection in cat_id_to_cat[annot["category_id"]]["skeleton"]:

                    from_point = connection[0]
                    to_point = connection[1]

                    if from_point in bad_points or to_point in bad_points:
                        continue

                    connections.append(
                        {
                            "id": index + 1,
                            "from": id_mapping[from_point],
                            "to": id_mapping[to_point],
                        }
                    )

                pointLabels = {}
                for kp_index, kp_name in enumerate(keypoint_names):
                    if kp_index + 1 in bad_points:
                        continue
                    pointLabels[id_mapping[kp_index + 1] - 1] = kp_name

                sa_obj = _create_vector_instance(
                    "template",
                    points,
                    pointLabels,
                    [],
                    cat_id_to_cat[annot["category_id"]]["supercategory"],
                    connections,
                    template_name=cat_id_to_cat[annot["category_id"]]["name"],
                )
                if str(annot["image_id"]) not in image_id_to_annotations:
                    image_id_to_annotations[str(annot["image_id"])] = [sa_obj]
                else:
                    image_id_to_annotations[str(annot["image_id"])].append(sa_obj)
        annotations_processed.append(annot)
    finish_event.set()
    tqdm_thread.join()
    save_sa_jsons(coco_json, image_id_to_annotations, output_dir)



"""
Dataloop to SA conversion method
"""
import json
import threading

from superannotate.logger import get_default_logger

from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_comment
from ..sa_json_helper import _create_sa_json
from ..sa_json_helper import _create_vector_instance
from .dataloop_helper import _create_attributes_list
from .dataloop_helper import _update_classes_dict

logger = get_default_logger()


def dataloop_to_sa(input_dir, task, output_dir):
    classes = {}
    json_data = list(input_dir.glob("*.json"))
    if task == "object_detection":
        instance_types = ["box"]
    elif task == "instance_segmentation":
        instance_types = ["segment"]
    elif task == "vector_annotation":
        instance_types = ["point", "box", "ellipse", "segment"]

    tags_type = "class"
    comment_type = "note"

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(json_data), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for json_file in json_data:
        dl_data = json.load(open(json_file))

        sa_metadata = {}
        if "itemMetadata" in dl_data and "system" in dl_data["itemMetadata"]:
            temp = dl_data["itemMetadata"]["system"]
            sa_metadata["name"] = temp["originalname"]
            sa_metadata["width"] = temp["width"]
            sa_metadata["height"] = temp["height"]

        sa_instances = []
        sa_tags = []
        sa_comments = []

        for ann in dl_data["annotations"]:
            if ann["type"] in instance_types:
                classes = _update_classes_dict(classes, ann["label"], ann["attributes"])

            attributes = _create_attributes_list(ann["attributes"])

            if ann["type"] in instance_types:
                if ann["type"] == "segment" and len(ann["coordinates"]) == 1:
                    points = []
                    for sub_list in ann["coordinates"]:
                        for sub_dict in sub_list:
                            points.append(sub_dict["x"])
                            points.append(sub_dict["y"])
                    instance_type = "polygon"
                elif ann["type"] == "box":
                    points = (
                        ann["coordinates"][0]["x"],
                        ann["coordinates"][0]["y"],
                        ann["coordinates"][1]["x"],
                        ann["coordinates"][1]["y"],
                    )
                    instance_type = "bbox"
                elif ann["type"] == "ellipse":
                    points = (
                        ann["coordinates"]["center"]["x"],
                        ann["coordinates"]["center"]["y"],
                        ann["coordinates"]["rx"],
                        ann["coordinates"]["ry"],
                        ann["coordinates"]["angle"],
                    )
                    instance_type = "ellipse"
                elif ann["type"] == "point":
                    points = (ann["coordinates"]["x"], ann["coordinates"]["y"])
                    instance_type = "point"
                sa_obj = _create_vector_instance(
                    instance_type, points, {}, attributes, ann["label"]
                )
                sa_instances.append(sa_obj)
            elif ann["type"] == comment_type:
                points = (
                    ann["coordinates"]["box"][0]["x"],
                    ann["coordinates"]["box"][0]["y"],
                )
                comments = []
                for note in ann["coordinates"]["note"]["messages"]:
                    comments.append({"text": note["body"], "email": note["creator"]})
                    sa_comment = _create_comment(points, comments)
                sa_comments.append(sa_comment)
            elif ann["type"] == tags_type:
                sa_tags.append(ann["label"])

        if "name" in sa_metadata:
            file_name = "%s___objects.json" % sa_metadata["name"]
        else:
            file_name = "%s___objects.json" % dl_data["filename"][1:]

        images_converted.append(file_name.replace("___objects.json ", ""))
        json_template = _create_sa_json(sa_instances, sa_metadata, sa_tags, sa_comments)
        write_to_json(output_dir / file_name, json_template)
    finish_event.set()
    tqdm_thread.join()
    return classes



"""
Supervisely to SA conversion method
"""
import json
import threading
from pathlib import Path

from superannotate.logger import get_default_logger

from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_sa_json
from ..sa_json_helper import _create_vector_instance
from .supervisely_helper import _base64_to_polygon
from .supervisely_helper import _create_attribute_list

logger = get_default_logger()


def supervisely_to_sa(json_files, class_id_map, task, output_dir):
    if task == "object_detection":
        instance_types = ["rectangle"]
    elif task == "instance_segmentation":
        instance_types = ["bitmap", "polygon"]
    elif task == "vector_annotation":
        instance_types = ["point", "rectangle", "line", "polygon", "cuboid", "bitmap"]

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(json_files), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for json_file in json_files:
        json_data = json.load(open(json_file))
        file_name = "%s___objects.json" % Path(json_file).stem
        sa_metadata = {
            "name": Path(json_file).stem,
            "width": json_data["size"]["width"],
            "height": json_data["size"]["height"],
        }

        sa_instances = []
        for obj in json_data["objects"]:
            if "classTitle" in obj and obj["classTitle"] in class_id_map.keys():
                attributes = []
                if "tags" in obj.keys():
                    attributes = _create_attribute_list(
                        obj["tags"], obj["classTitle"], class_id_map
                    )

                if obj["geometryType"] in instance_types:
                    if obj["geometryType"] == "point":
                        points = (
                            obj["points"]["exterior"][0][0],
                            obj["points"]["exterior"][0][1],
                        )
                        instance_type = "point"
                    elif obj["geometryType"] == "line":
                        instance_type = "polyline"
                        points = [
                            item for el in obj["points"]["exterior"] for item in el
                        ]
                    elif obj["geometryType"] == "rectangle":
                        instance_type = "bbox"
                        points = (
                            obj["points"]["exterior"][0][0],
                            obj["points"]["exterior"][0][1],
                            obj["points"]["exterior"][1][0],
                            obj["points"]["exterior"][1][1],
                        )
                    elif obj["geometryType"] == "polygon":
                        instance_type = "polygon"
                        points = [
                            item for el in obj["points"]["exterior"] for item in el
                        ]
                    elif obj["geometryType"] == "cuboid":
                        instance_type = "cuboid"
                        points = (
                            obj["points"][0][0],
                            obj["points"][0][1],
                            obj["points"][2][0],
                            obj["points"][2][1],
                            obj["points"][4][0],
                            obj["points"][4][1],
                            obj["points"][5][0],
                            obj["points"][6][1],
                        )
                    elif obj["geometryType"] == "bitmap":
                        for ppoints in _base64_to_polygon(obj["bitmap"]["data"]):
                            points = [
                                x + obj["bitmap"]["origin"][0]
                                if i % 2 == 0
                                else x + obj["bitmap"]["origin"][1]
                                for i, x in enumerate(ppoints)
                            ]
                        instance_type = "polygon"

                    sa_obj = _create_vector_instance(
                        instance_type, points, {}, attributes, obj["classTitle"]
                    )
                    sa_instances.append(sa_obj)
        images_converted.append(Path(json_file).stem)
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)
    finish_event.set()
    tqdm_thread.join()


def supervisely_keypoint_detection_to_sa_vector(
    json_files, class_id_map, meta_json, output_dir
):
    classes_skeleton = {}
    for class_ in meta_json["classes"]:
        if class_["shape"] != "graph":
            continue

        classes_skeleton[class_["title"]] = {"edges": [], "nodes": {}}
        nodes = class_["geometry_config"]["nodes"]
        for node, value in nodes.items():
            classes_skeleton[class_["title"]]["nodes"][node] = value["label"]

        edges = class_["geometry_config"]["edges"]
        for edge in edges:
            classes_skeleton[class_["title"]]["edges"].append(
                (edge["src"], edge["dst"])
            )

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(json_files), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for json_file in json_files:
        file_name = "%s___objects.json" % (Path(json_file).stem)
        json_data = json.load(open(json_file))
        sa_metadata = {
            "name": Path(json_file).stem,
            "width": json_data["size"]["width"],
            "height": json_data["size"]["height"],
        }
        sa_instances = []

        for obj in json_data["objects"]:
            if "classTitle" in obj and obj["classTitle"] in class_id_map.keys():
                attributes = []
                if "tags" in obj.keys():
                    attributes = _create_attribute_list(
                        obj["tags"], obj["classTitle"], class_id_map
                    )

                    if obj["geometryType"] == "graph":
                        good_nodes = []
                        nodes = obj["nodes"]
                        index = 1
                        points = []
                        pointLabels = {}
                        for node, value in nodes.items():
                            good_nodes.append(node)
                            points.append(
                                {
                                    "id": index,
                                    "x": value["loc"][0],
                                    "y": value["loc"][1],
                                }
                            )
                            pointLabels[index - 1] = classes_skeleton[
                                obj["classTitle"]
                            ]["nodes"][node]
                            index += 1

                        index = 1
                        connections = []
                        for edge in classes_skeleton[obj["classTitle"]]["edges"]:
                            if edge[0] not in good_nodes or edge[1] not in good_nodes:
                                continue

                            connections.append(
                                {
                                    "id": index,
                                    "from": good_nodes.index(edge[0]) + 1,
                                    "to": good_nodes.index(edge[1]) + 1,
                                }
                            )
                            index += 1
                        sa_obj = _create_vector_instance(
                            "template",
                            points,
                            pointLabels,
                            attributes,
                            obj["classTitle"],
                            connections,
                        )
                        sa_instances.append(sa_obj)
        images_converted.append(Path(json_file).stem)
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)
    finish_event.set()
    tqdm_thread.join()



import copy
from abc import ABCMeta
from abc import abstractmethod
from typing import Any
from typing import Type

from pydantic import BaseModel
from pydantic import Extra


class BaseValidator(metaclass=ABCMeta):
    MODEL: BaseModel()

    def __init__(self, data: Any, allow_extra: bool = True):
        self.data = data
        self._validation_output = None
        self._extra = Extra.allow if allow_extra else Extra.forbid

    @classmethod
    def validate(cls, data: Any, extra=True):
        return cls.MODEL(**data)

    def _validate(self):
        model = copy.deepcopy(self.MODEL)
        model.Config.extra = self._extra
        self.data = model(**self.data).dict(by_alias=True, exclude_none=True)

    @abstractmethod
    def is_valid(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def generate_report(self) -> str:
        raise NotImplementedError


class BaseAnnotationValidator(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def get_pixel_validator() -> Type[BaseValidator]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_vector_validator() -> Type[BaseValidator]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_video_validator() -> Type[BaseValidator]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_document_validator() -> Type[BaseValidator]:
        raise NotImplementedError
